@ECHO OFF
REM ==========================================================
REM charm.bat -- a script to create an ARM binary or library =
REM                                                          =
REM Call as:                                                 =
REM                                                          =
REM   charm filelist                                         =
REM                                                          =
REM where filelist specifies one or more source (.c) files   =
REM files, which are either single names or names using      =
REM wildcards, for example: charm master.c foo*.c            =
REM                                                          =
REM   returns 0 if result OK, non-zero if some error         =
REM                                                          =
REM The following optional environment variables may be set  =
REM before calling this script (note that these are case     =
REM sensitive):                                              =
REM                                                          =
REM Usually required:                                        =
REM                                                          =
REM   libs=<list>    -- list of sibling library directory    =
REM                     names to search for .h files, .ld    =
REM                     link scripts, and .a libraries       =
REM                     (<lib>\public for .h files, too)     =
REM   script=<name>  -- name of the .ld file (processor link =
REM                     script); not required if lib=1       =
REM                                                          =
REM Optional:                                                =
REM                                                          =
REM   cextra=<list>  -- extra flags to pass for compilation  =
REM   hex=1          -- write a .hex rather than .bin (if    =
REM                     lib is not 1)                        =
REM   lax=1          -- use less-strict C compiler checking  =
REM   ldextra=<list> -- extra flags to pass for link         =
REM   lib=1          -- build a library (.a) rather than a   =
REM                     binary or hexadecimal output file    =
REM   map=1          -- create a .map file with crossrefs    =
REM                     (if building a binary)               =
REM   ocextra=<list> -- extra flags to pass for objcopy      =
REM                                                          =
REM Notes:                                                   =
REM   1. The search for .h files includes the current        =
REM      directory and all directories on the libs list,     =
REM      and any public directory in these.                  =
REM                                                          =
REM   2. The output file name will be the same as the name   =
REM      of the current directory.                           =
REM                                                          =
REM   3. Tollos directories are all siblings (at the same    =
REM      level), and must not have blanks in their names.    =
REM                                                          =
REM   4. If the output file cannot be created it is erased   =
REM      (so a check for its existence will determine if     =
REM      any failure occurred).                              =
REM                                                          =
REM   5. .o files matching the source file(s) are erased.    =
REM                                                          =
REM   6. For extra reliability, set unused variables listed  =
REM      above to empty [SET foobar=]                        =
REM ==========================================================
REM Author: Mike Cowlishaw -- use at your own risk           =
REM ==========================================================

REM rc is the return code; set to 1 initially to assume the worst
SET rc=1
IF .%1==. GOTO noarg

SETLOCAL enabledelayedexpansion

REM Get name of current directory
FOR %%d IN (.) DO (SET thisdir=%%~nd)

REM Build the libraries list and search lists
SET libfiles=
SET includes=-I.  -I.\public
SET linkdirs=-L .
FOR %%b IN (%libs%) DO (
  SET libfiles=!libfiles! ..\%%~nb\%%~nb.a
  SET includes=!includes! -I..\%%~nb -I..\%%~nb\public
  SET linkdirs=!linkdirs! -L ..\%%~nb
  )

REM Decide output file and formats, and quietly delete if exist
IF .%lib%==.1 GOTO reqlib
  IF .%hex%==.1 GOTO reqhex
    SET finfile=%thisdir%.bin
    SET objform=binary
    GOTO reqcont
  :reqhex
    SET finfile=%thisdir%.hex
    SET objform=ihex
    GOTO reqcont
:reqlib
  SET finfile=%thisdir%.a
:reqcont
SET outfile=%thisdir%.out
ERASE %finfile% 2>nul
ERASE %outfile% 2>nul

REM ==========================================================
REM Compile the .c file(s)                                   =
REM ==========================================================

SET flags=-c -fno-common -Wall -g -mcpu=cortex-m3 -mthumb
SET flags=%flags% -ffunction-sections -Os
IF .%lax%==.1 GOTO laxskip
  SET flags=%flags% -std=c99 -pedantic -Wextra
  SET flags=%flags% -Wstrict-prototypes -Wnested-externs -Wmissing-prototypes
  SET flags=%flags% -Wcast-qual -fstrict-aliasing -Wstrict-aliasing
:laxskip
SET compile=arm-none-eabi-gcc %flags% %cextra% %includes%

SET objects=
FOR %%c IN (%*) DO (
  %compile% %%~c
  IF NOT ERRORLEVEL 1 (ECHO %%~c Compiled OK)  ELSE GOTO quit
  SET objects=!objects! .\%%~nc.o
  )

REM ==========================================================
REM Build the library or binary                              =
REM ==========================================================
IF .%lib%==.1 GOTO makelib
  IF .%script%==. GOTO nolink
  SET ldflags=-nostartfiles -static --gc-sections -o %outfile%
  SET ldscript=%linkdirs% -T %script%.ld

  IF NOT .%map%==.1 GOTO nomap
    SET mapfile=%thisdir%.map
    ERASE mapfile 2>nul
    SET ldflags=%ldflags% -Map=%mapfile% --cref
  :nomap

  REM duplicate libfiles to allow mutual self-reference
  SET libfiles=%libfiles% %libfiles%

  arm-none-eabi-ld %ldflags% %ldscript% %objects% %libfiles% %ldextra%
  IF NOT ERRORLEVEL 1 (ECHO Linked OK)  ELSE GOTO quit
  arm-none-eabi-objcopy -O%objform% %ocextra% %outfile% %finfile%
  IF NOT ERRORLEVEL 1 GOTO done
  GOTO quit

:makelib
  arm-none-eabi-ar cr %finfile% %objects%
  IF NOT ERRORLEVEL 1 GOTO done
  GOTO quit


REM ==========================================================
REM Exit paths                                               =
REM ==========================================================
:nolink
ECHO No link script was set before calling %~f0
ECHO e.g., use: SET script=lpc1768
GOTO quit

:noarg
ECHO No parameter (e.g., sample.c) was passed to %~f0
GOTO quit

:done
FOR %%f IN (%finfile%) DO SET size=%%~zf
ECHO %finfile% created OK, size %size% bytes
SET rc=0

:quit
REM Quietly delete any debris
FOR %%o IN (%objects%) DO ERASE %%o 2>nul
ERASE %outfile% 2>nul
EXIT /B %rc%
