/* --------------------------------------------------------------- */
/* miniDec.c -- a decimal package for small systems                */
/* --------------------------------------------------------------- */
/*                                                                 */
/* Copyright (c) Mike Cowlishaw, 2010-2011.  All rights reserved.  */
/*                                                                 */
/* Permission to use, copy, modify, and/or distribute this         */
/* software for any purpose with or without fee is hereby granted, */
/* provided that the above copyright notice and this permission    */
/* notice appear in all copies, and that notice and the date of    */
/* any modifications be added to the software.                     */
/*                                                                 */
/* This software is provided "as is".  No warranties, whether      */
/* express, implied, or statutory, including, but not limited to,  */
/* implied warranties of merchantability and fitness for a         */
/* particular purpose apply to this software.  The author shall    */
/* not, in any circumstances, be liable for special, incidental,   */
/* or consequential damages, for any reason whatsoever.            */
/*                                                                 */
/* --------------------------------------------------------------- */
/*                                                                 */
/* This package supports a simple decimal format, intended for the */
/* storage, management, and display of decimal data.  As such, it  */
/* provides conversions to and from ints, doubles, and strings in  */
/* various forms, along with decQuantize to 'shape' decimal data.  */
/*                                                                 */
/* Arithmetic operations are currently not included, but the       */
/* bounds of the format conform to IEEE 754 rules (section 3) and  */
/* IEEE 754 arithmetic functions could be added later.             */
/*                                                                 */
/* Overflows or bad conversions result in a Infinities or quiet    */
/* NaNs (decNaN), as appropriate.  For details not covered in the  */
/* function descriptions, see the General Decimal Arithmetic       */
/* Specification.                                                  */
/*                                                                 */
/* Decimal numbers are passed by copy since they are smaller than  */
/* two words (assuming a 32-bit system).                           */
/*                                                                 */
/* The maximum coefficient is 9 digits and emax is 999.            */
/*                                                                 */
/* --------------------------------------------------------------- */
// Implementation notes:
//
//  1. Emphasis here is more on simplicity and small size rather
//     than function or speed; hence only one rounding mechanism
//     is used (round half up); this could be extended later, using
//     conditional compilation, or special functions, e.g.,
//     decQuantizeMode.
//
//  2. Code follows IEEE 754 arithmetic rules (but not any
//     standard interchange format).
//
//  3. This code only inspects or changes the fields of a dec
//     through macros.
//
//  4. There is no concept of 'context', e.g., IEEE 754 flags.
//
//  5. miniDec does not fold values with exponents near emax
//     (e.g., 1E+999 is stored as [1, 999], not [100000000, 991]).
//
//  6. The notation ** means 'raise to the power'.
//
//  7. This file and block comments conform to Tollos coding
//     and documentation conventions (see speleotrove.com/tollos)
// ---------------------------------------------------------------

#include "miniDec.h"
#include <ctype.h>
#include <math.h>
#include <stddef.h>

// shorthand types (int assumed 32; asserted in .h)
typedef uint32_t  uint;
// ensure 'int' is 32-bit signed
#if (INT_MAX != 2147483647)
  #error miniDec assumes 'int' is 32 bits, and signed
#endif

// Constants
const dec decZero =DECINITZERO;    // 0
const dec decZeroM=DECINITZEROM;   // -0
const dec decNaN  =DECINITQNAN;
const dec decSNaN =DECINITSNAN;
const dec decInf  =DECINITINF;     // infinity
const dec decInfM =DECINITINFM;    // -infinity

// Powers of ten (powers[n]==10**n, 0<=n<=9)
#define MILLION 1000000ULL
const int powers[10]={1, 10, 100, 1000, 10000, 100000, MILLION,
                      MILLION*10, MILLION*100, MILLION*1000};

// Macros

// simple extracts
#define getCoeff(d) ((int)(d).coef)
#define getExp(d)   ((int)((d).meta>>4))
#define getSign(d)  ((int)(((d).meta & DECSIGN)!=0))
// simple tests
#define isFinite(d)  (((d).meta & (DECQNAN | DECSNAN | DECINF))==0)
#define isInf(d)     (((d).meta & DECINF)!=0)
#define isNaN(d)     (((d).meta & (DECQNAN | DECSNAN))!=0)
#define isQNaN(d)    (((d).meta & DECQNAN)!=0)
#define isSigned(d)  (((d).meta & DECSIGN)!=0)
#define isSNaN(d)    (((d).meta & DECSNAN)!=0)
#define isSpecial(d) (((d).meta & (DECQNAN | DECSNAN | DECINF))!=0)
#define isZero(d)    (((d).meta & (DECQNAN | DECSNAN | DECINF))?0:(d).coef==0)
// simple sets [note first parameter is a local name, not address]
#define setCoeff(d, coe) ((d).coef=(coe))
#define setExp(d, exp)   ((d).meta=((d).meta & 0x0f) | ((exp)<<4))
#define setSign(d, sign) ((d).meta=((d).meta&~DECSIGN) \
                               | ((sign) ? DECSIGN : 0))
// two decimals
#define identical(d1, d2) ((d1).meta==(d2).meta && (d1).coef==(d2).coef)

// Local functions
static int streq(const char *char1, const char *char2);

// Assertions
#if DECDIGITS!=9
  #if DECDIGITS>9
    #error DECDIGITS too large (>9)
  #else
    #error DECDIGITS is not 9 (coefficient handling needs changing)
  #endif
#endif

// Functions are ordered alphabetically

/* --------------------------------------------------------------- */
/* decFromDouble -- double to decimal                              */
/*                                                                 */
/*   dub -- double (IEEE 754 binary64) to convert                  */
/*                                                                 */
/*   returns decimal conversion of dub                             */
/*                                                                 */
/* Finite numbers will always be in range, but will be rounded to  */
/* 9 decimal digits (or one, for zeros).  All NaNs return decNaN.  */
/* --------------------------------------------------------------- */
#define getDubExp(uu) ((((uu)>>52)&0x7ff)-1023) // get binary exponent
dec decFromDouble(double dub) {
  int bexp;                             // binary exponent
  int dexp;                             // decimal exponent
  int coeff;                            // decimal integer coefficient
  dec ans;                              // for result
  int div, residue;                     // work for rounding
  union {                               // double-mapper (non-endian)
    uint64_t uu;
    double   dd;
    } map;

  map.dd=dub;                           // map.uu is now dub as an integer
  bexp=getDubExp(map.uu);               // extract 11-bit signed exponent

  if (bexp==0x400) {                    // specials are easy
    if ((map.uu&0x000fffffffffffffULL)!=0) return decNaN;
    if (map.uu>>63) return decInfM;
    return decInf;
    }
  if (map.uu>>63) ans=decZeroM;         // prepare result with sign bit
   else ans=decZero;
  if (bexp==-1023) {                    // zero or subnormal
    uint32_t u;                         // work
    if (map.uu<<1==0) return ans;       // quick exit for zeros
    // subnormal; count leading zeros in the significant to calculate
    // true bexp -- do this in parts to avoid long shifts
    u=(uint32_t)(map.uu>>32)<<12;       // top (shifted to MSB)
    if (u==0) {                         // first 1 is in lower part
      bexp-=20;                         // 20 leading zeros already
      u=map.uu;                         // search down this
      }
    for (; (int)u>0; u<<=1) bexp--;     // at end, bexp >= -1074
    }

  // choose decimal exponent so 2**bexp/10**dexp is always >=1 and <10
  dexp=(315652*bexp)>>20;               // or 30103*bexp/100000

  // calculate dd = (dub/10**dexp) * (10**16/2**52), which will
  // leave a sixteen-digit integer decimal coefficient in dd;
  // use two multiplies if pow(10, 16-dexp) would overflow to inf
  if (16-dexp<=308) map.dd=dub*pow(10.0, 16-dexp);  // dexp is >= -292
   else map.dd=1.0E+308*dub*pow(10.0, 16-dexp-308); // dexp is >= -324
  // now 1E+16<=map.dd<44.3E+16, so 52<=bexp(map.uu)<=58; map.uu is integer

  bexp=getDubExp(map.uu);               // extract 11-bit signed exponent
  map.uu&=0x000fffffffffffffULL;        // drop exponent and sign
  map.uu|=0x0010000000000000ULL;        // add implied bit
  map.uu<<=(bexp-52);                   // adjust so bexp would be 52

  // this leaves map.uu with either 17 (common) or 18 decimal digits;
  // round this to 9 digits
  div=MILLION*100;                      // assume will drop 8 digits
  if (map.uu>=MILLION*MILLION*100000) { // 18 digits [ULL constant]
    div*=10;                            // will lop off 9
    dexp++;                             // adjust exponent to match
    }
  coeff=(int)(map.uu/div);              // shift
  residue=(int)(map.uu%div);            // remainder
  if (residue>=div/2) {                 // round up
    if (coeff!=999999999) coeff++;      // simple bump
     else {                             // the infamous case
      coeff=100000000;                  // [yes, it can happen]
      dexp++;
      }
    }
  // now 9 digits; the range can neither underflow nor overflow
  setExp(ans, dexp+8-16);               // +8 for rounding and * 1E-16
  setCoeff(ans, coeff);                 // copy coefficient
  return ans;
  } // decFromDouble

/* --------------------------------------------------------------- */
/* decFromInt -- int to decimal                                    */
/*                                                                 */
/*   i -- int to convert [DECMINCOE <= i <= DECMAXCOE]             */
/*                                                                 */
/*   returns decimal with the value of the int, or decNaN if out   */
/*     of range                                                    */
/* --------------------------------------------------------------- */
dec decFromInt(int i) {
  dec ans;                    // target
  if (i>=0) {
    if (i>DECMAXCOE) return decNaN;
    ans=decZero;
    }
   else {
    if (i<DECMINCOE) return decNaN;
    i=-i;                     // now 0 or positive
    ans=decZeroM;
    }
  setCoeff(ans, i);           // add coefficient
  return ans;
  } // decFromInt

/* --------------------------------------------------------------- */
/* decFromInts -- int coefficent and exponent to decimal           */
/*                                                                 */
/*   coe -- coefficient to convert [DECMINCOE <= coe <= DECMAXCOE] */
/*   exp -- exponent to use                                        */
/*                                                                 */
/*   returns decimal constructed from coe and exp, or decNaN if    */
/*     the resulting number would be out of range                  */
/*                                                                 */
/* The number is in range if it is zero, or it is non-zero and its */
/* magnitude is in the range 0.00000001*10**-998 through           */
/* 9.99999999*10**999, provided that the exponent needed to        */
/* represent it exactly with the coefficient as an integer         */
/* (including trailing zeros) is in the range -1006 through 999.   */
/* --------------------------------------------------------------- */
dec decFromInts(int coe, int exp) {
  dec ans=decFromInt(coe);
  if (!isNaN(ans)) {          // so far, so good ..
    int sigs=decGetDigits(ans);
    // check exponent range
    if (exp+DECDIGITS-1<DECEMIN
     || exp+sigs-1>DECEMAX) return decNaN;
    setExp(ans, exp);
    }
  return ans;
  } // decFromInts

/* --------------------------------------------------------------- */
/* decFromString -- string to decimal                              */
/*                                                                 */
/*   str -- string to convert                                      */
/*                                                                 */
/*   returns decimal holding the converted string                  */
/*                                                                 */
/* If the coefficient has more than 9 digits, or the number        */
/* represented by the string is out of range, or the syntax is     */
/* invalid, then a quiet NaN is returned.  No rounding takes       */
/* place; for finite numbers this conversion is exact if it        */
/* succeeds.                                                       */
/*                                                                 */
/* The number is in range if it is zero, or it is non-zero and its */
/* magnitude is in the range 0.00000001*10**-998 through           */
/* 9.99999999*10**999, provided that the exponent needed to        */
/* represent it exactly with the coefficient as an integer         */
/* (including trailing zeros) is in the range -1006 through 999.   */
/*                                                                 */
/* The spellings for the special values are as recommended by IEEE */
/* 754, and are independent of case; the following are accepted:   */
/* "infinity", "inf", "nan", or "snan".                            */
/*                                                                 */
/* See the General Arithmetic Specification for a formal           */
/* description of allowed syntax.                                  */
/* --------------------------------------------------------------- */
dec decFromString(const char * str) {
  int exp=0, sign=0, coef;              // parts of a number
  int digits, sigs;                     // work
  const char *c, *dotchar=NULL, *cfirst=str;  // ..
  dec ans;                              // ..

  // detect and validate the coefficient, including any leading,
  // trailing, or embedded '.'
  for (c=str; *c; c++) {                // -> input character
    if (*c>='0' && *c<='9') continue;   // good digit
    if (*c=='.' && dotchar==NULL) {
      dotchar=c;                        // record offset into decimal part
      continue;}
    if (c==str) {                       // first in string...
      if (*c=='-') {                    // valid - sign
        cfirst++;
        sign=1;
        continue;}
      if (*c=='+') {                    // valid + sign
        cfirst++;
        continue;}
      }
    // *c is not a digit, terminator, or a valid +, -, or '.'
    break;                              // (E, or start of special)
    } // c loop

  digits=c-cfirst;                      // digits (+1 if a dot)

  if (digits>0) {                       // had digits and/or dot
    // should be processing a finite number
    const char *clast=c-1;              // note last coefficient char position
    int  expadj;                        // exponent adjustment

    if (dotchar==NULL) expadj=0;
     else {                             // had a '.'
      digits--;                         // remove from digits count
      if (digits==0) return decNaN;     // was dot alone: bad syntax
      expadj=(int)(clast-dotchar);      // exponent adjustment
      // [the '.' can now be ignored]
      }

    if (*c!='\0') {                     // something follows the coefficient
      // had some digits, and more characters; E[+|-]nnn expected
      if (*c=='E' || *c=='e') c++;
       else return decNaN;
      if (*c=='-' || *c=='+') c++;      // step over sign [c=clast+2]
      if (*c=='\0') return decNaN;      // no digits

      // collect exponent digits (arbitrary limit here of 4 sig digits)
      for (;; c++) {
        if (*c<'0' || *c>'9') break;    // not '0' through '9'
        exp=exp*10+(*c-'0');            // accumulate
        if (exp>9999) return decNaN;    // too big
        }
      if (*c!='\0') return decNaN;      // junk on end
      if (*(clast+2)=='-') exp=-exp;    // was negative
      // exact range is checked below, when coefficent digits are known
      } // exponent part

    exp-=expadj;

    // collect the coefficient; leading zeros can be ignored, but
    // only 9 significant digits are allowed
    coef=0;                             // coefficient
    sigs=0;                             // no significant digit yet
    for (c=cfirst; c<=clast; c++) {
      if (*c=='.') continue;            // ignore
      if (sigs>0) {
        if (sigs==9) return decNaN;     // too long
        sigs++;
        }
       else {                           // sigs==0
        if (*c=='0') continue;          // leading 0
        sigs=1;                         // first significant digit
        }
      coef=coef*10+((uint)*c-(uint)'0'); // accumulate
      }
    if (sigs==0) sigs=1;                // all zeros

    // check exponent range
    if (exp+DECDIGITS-1<DECEMIN || exp+sigs-1>DECEMAX) return decNaN;
    ans=decZero;                        // [placate compiler]
    setCoeff(ans, coef);                // construct the number
    setExp(ans, exp);                   // ..
    setSign(ans, sign);                 // ..
    // decPrint(">>", ans);
    return ans;
    } // finite

  // only Infinities and NaNs are allowed, here; all are allowed sign
  if (streq(c, "infinity") || streq(c, "inf")) ans=decInf;
   else if (streq(c, "nan")) ans=decNaN;
   else if (streq(c, "snan")) ans=decSNaN;
   else return decNaN;                  // includes empty case
  setSign(ans, sign);
  return ans;
  } // decFromString

/* --------------------------------------------------------------- */
/* decGetCoeff -- return the coefficient of a decimal as an int    */
/*                                                                 */
/*   num -- decimal to process                                     */
/*                                                                 */
/*   returns the unsigned coefficient of num, as an int            */
/*                                                                 */
/* The result will be in the range 0 .. 999999999                  */
/* --------------------------------------------------------------- */
int decGetCoeff(dec num) {
  return getCoeff(num);
  } // decGetCoeff

/* --------------------------------------------------------------- */
/* decGetDigits -- return the number of digits in a decimal        */
/*                                                                 */
/*   num -- decimal to check                                       */
/*                                                                 */
/*   returns the number of significant digits in the coefficient   */
/*     of num (or returns 1 if the coefficient is 0, or num is     */
/*     infinite or a NaN)                                          */
/* --------------------------------------------------------------- */
int decGetDigits(dec num) {
  int i;                      // work
  if (getCoeff(num)==0) return 1;
  // test upwards because most numbers are small
  for (i=1; i<9; i++) if (getCoeff(num)<powers[i]) break;
  return i;
  } // decGetDigits

/* --------------------------------------------------------------- */
/* decGetExp -- return the exponent of a decimal as an int         */
/*                                                                 */
/*   num -- decimal to process                                     */
/*                                                                 */
/*   returns the signed exponent of num                            */
/*                                                                 */
/* The result will be in the range -1006 .. 999                    */
/* --------------------------------------------------------------- */
int decGetExp(dec num) {
  return getExp(num);
  } // decGetExp

/* --------------------------------------------------------------- */
/* decGetSign -- return the sign of a decimal as an int            */
/*                                                                 */
/*   num -- decimal to process                                     */
/*                                                                 */
/*   returns the sign of num                                       */
/*                                                                 */
/* The result will be 1 if num is signed or 0 if it is not.        */
/* The sign indicates negative for non-zero numbers and infinity.  */
/* --------------------------------------------------------------- */
// This is identical to decIsSigned(num)
int decGetSign(dec num) {
  return getSign(num);
  } // decGetSign

/* --------------------------------------------------------------- */
/* decIdentical -- test if two decimals are identical              */
/*                                                                 */
/*   num1 -- first decimal to test                                 */
/*   num2 -- second decimal to test                                */
/*                                                                 */
/*   returns 1 the two decimals are identical in all respects      */
/*                                                                 */
/* Decimals are identical if they have the same sign, and also if  */
/* finite they have the same exponent and coefficient, or if not   */
/* finite they are the same kind of NaN or they are both infinite. */
/* Note that decimals that are not identical might be numerically  */
/* equal (for example, 2.50 is numerically equal to 2.5).          */
/* --------------------------------------------------------------- */
int decIdentical(dec num1, dec num2) {
  return identical(num1, num2);
  } // decIdentical

/* --------------------------------------------------------------- */
/* decIsFinite -- test if a decimal is finite                      */
/*                                                                 */
/*   num -- decimal to test                                        */
/*                                                                 */
/*   returns 1 if num is neither a NaN nor an Infinity,            */
/*     0 in all other cases.                                       */
/* --------------------------------------------------------------- */
int decIsFinite(dec num) {
  return isFinite(num);
  } // decIsFinite

/* --------------------------------------------------------------- */
/* decIsInf -- test if a decimal is infinite                       */
/*                                                                 */
/*   num -- decimal to test                                        */
/*                                                                 */
/*   returns 1 if num is an Infinity,                              */
/*     0 in all other cases.                                       */
/* --------------------------------------------------------------- */
int decIsInf(dec num) {
  return isInf(num);
  } // decIsInf

/* --------------------------------------------------------------- */
/* decIsNaN -- test if a decimal is a NaN                          */
/*                                                                 */
/*   num -- decimal to test                                        */
/*                                                                 */
/*   returns 1 if num is a qNaN or an sNaN,                        */
/*     0 in all other cases.                                       */
/* --------------------------------------------------------------- */
int decIsNaN(dec num) {
  return isNaN(num);
  } // decIsNaN

/* --------------------------------------------------------------- */
/* decIsNegative -- test if a decimal is a number and negative     */
/*                                                                 */
/*   num -- decimal to test                                        */
/*                                                                 */
/*   returns 1 if num is a number (finite or infinite) and         */
/*     negative (<0),                                              */
/*     0 in all other cases.                                       */
/* --------------------------------------------------------------- */
int decIsNegative(dec num) {
  if (isNaN(num) || isZero(num)) return 0;   // NaN or 0
  return isSigned(num);                      // sign=1?
  } // decIsNegative

/* --------------------------------------------------------------- */
/* decIsPositive -- test if a decimal is a number and positive     */
/*                                                                 */
/*   num -- decimal to test                                        */
/*                                                                 */
/*   returns 1 if num is a number (finite or infinite) and         */
/*     positive (>0),                                              */
/*     0 in all other cases.                                       */
/* --------------------------------------------------------------- */
int decIsPositive(dec num) {
  if (isNaN(num) | isZero(num)) return 0;    // NaN or 0
  return !isSigned(num);                     // sign=0?
  } // decIsPositive

/* --------------------------------------------------------------- */
/* decIsQNaN -- test if a decimal is a quiet NaN                   */
/*                                                                 */
/*   num -- decimal to test                                        */
/*                                                                 */
/*   returns 1 if num is a qNaN,                                   */
/*     0 in all other cases.                                       */
/* --------------------------------------------------------------- */
int decIsQNaN(dec num) {
  return isQNaN(num);
  } // decIsQNaN

/* --------------------------------------------------------------- */
/* decIsSigned -- test if a decimal has a sign bit of 1            */
/*                                                                 */
/*   num -- decimal to test                                        */
/*                                                                 */
/*   returns 1 if the sign bit of num is 1, and 0 if it is not.    */
/*                                                                 */
/* This may be used to test the sign of a zero or a NaN.           */
/* --------------------------------------------------------------- */
int decIsSigned(dec num) {
  return isSigned(num);
  } // decIsSigned

/* --------------------------------------------------------------- */
/* decIsSNaN -- test if a decimal is signalling NaN                */
/*                                                                 */
/*   num -- decimal to test                                        */
/*                                                                 */
/*   returns 1 if num is an sNaN,                                  */
/*     0 in all other cases.                                       */
/* --------------------------------------------------------------- */
int decIsSNaN(dec num) {
  return isSNaN(num)!=0;
  } // decIsSNaN

/* --------------------------------------------------------------- */
/* decIsSpecial -- test if a decimal is a NaN or Infinity          */
/*                                                                 */
/*   num -- decimal to test                                        */
/*                                                                 */
/*   returns 1 if num is not a finite number,                      */
/*     0 in all other cases.                                       */
/* --------------------------------------------------------------- */
int decIsSpecial(dec num) {
  return isSpecial(num);
  } // decIsSpecial

/* --------------------------------------------------------------- */
/* decIsZero -- test if a decimal is a number with value 0         */
/*                                                                 */
/*   num -- decimal to test                                        */
/*                                                                 */
/*   returns 1 if num is a finite number with a value of 0,        */
/*     0 in all other cases.                                       */
/*                                                                 */
/* This test is independent of the sign and exponent of num.       */
/* --------------------------------------------------------------- */
int decIsZero(dec num) {
  return isZero(num);
  } // decIsZero

/* --------------------------------------------------------------- */
/* decQuantize -- round or extend a decimal to match another       */
/*                                                                 */
/*   a -- decimal to be modified                                   */
/*   b -- pattern to match                                         */
/*                                                                 */
/*   returns the decimal with the closest value to a which has the */
/*     exponent of b                                               */
/*                                                                 */
/* If either value is a NaN then a quiet NaN is returned.  If both */
/* are infinite, then a is returned.                               */
/*                                                                 */
/* Otherwise, the returned value will have the same exponent as a; */
/* if this is not possible (not enough digits, or either a or b is */
/* infinite) then this is an error and a quiet NaN is returned.    */
/* If rounding is needed then round half up is used.               */
/* --------------------------------------------------------------- */
dec decQuantize(dec a, dec b) {
  int expa, expb;                  // exponents
  int adjust;                      // exponent adjustment needed
  int diga;                        // digits
  int residue;                     // rounding remainder

  if (isSpecial(a) || isSpecial(b)) {   // handle all specials
    if (isInf(a) && isInf(b)) return a; // b must be Inf
    // any NaNs (or mixed ninf/non-inf, which is invalid) return NaN
    return decNaN;
    }

  // both are finite
  expb=getExp(b);
  if (isZero(a)) {                 // easy
    setExp(a, expb);               // set a's exponent to b's
    return a;
    }
  expa=getExp(a);
  // non-zero coefficient
  adjust=expb-expa;
  if (adjust==0) return a;         // nothing to do
  diga=decGetDigits(a);
  setExp(a, expb);                 // set the exponent optimistically
  // the sign of a is always unchanged

  if (adjust<0) {                  // padding with -adjust zeros required
    if (diga-adjust>DECDIGITS) return decNaN;     // cannot fit
    setCoeff(a, getCoeff(a)*powers[-adjust]);     // shift left
    return a;                      // done
    }

  // shifting adjust digits to the right; rounding needed
  if (adjust>=DECDIGITS) {         // large shift
    setCoeff(a, 0);
    return a;
    }

  // shift 1..9
  residue=getCoeff(a)%powers[adjust];        // remainder
  setCoeff(a , getCoeff(a)/powers[adjust]);  // shift
  if (residue>=powers[adjust]/2) setCoeff(a, getCoeff(a)+1); // round
  return a;
  } // decQuantize

/* --------------------------------------------------------------- */
/* decSetCoeff -- set the coefficient of a decimal from an int     */
/*                                                                 */
/*   num -- target decimal                                         */
/*   coe -- int to use to set the coefficient (0 .. 999999999)     */
/*                                                                 */
/* The coefficient of the decimal will be set; no checking is      */
/* carried out so it must be in range.                             */
/* This does not set or change the sign or exponent of num.        */
/* --------------------------------------------------------------- */
void decSetCoeff(dec *num, int coe) {
  setCoeff(*num, coe);
  return;
  } // decSetCoeff

/* --------------------------------------------------------------- */
/* decSetExp -- set the exponent of a decimal                      */
/*                                                                 */
/*   num -- target decimal                                         */
/*   exp -- sign to be set (-1006 .. 991)                          */
/*                                                                 */
/* The exponent of the decimal will be set; no checking is carried */
/* out so it must be in range.                                     */
/* This does not set or change the sign or coefficient of num.     */
/* --------------------------------------------------------------- */
void decSetExp(dec *num, int exp) {
  setExp(*num, exp);
  return;
  } // decSetExp

/* --------------------------------------------------------------- */
/* decSetSign -- set the sign of a decimal                         */
/*                                                                 */
/*   num  -- target decimal                                        */
/*   sign -- sign to be set (0 or 1)                               */
/*                                                                 */
/* The sign of the decimal will be set to 0 if sign is 0, or 1 if  */
/* sign has any other value.                                       */
/* This does not set or change the coefficient or exponent of num. */
/* --------------------------------------------------------------- */
void decSetSign(dec *num, int sign) {
  setSign(*num, sign);
  return;
  } // decSetSign

/* --------------------------------------------------------------- */
/* decToDouble -- convert a decimal to a double                    */
/*                                                                 */
/*   num -- decimal to convert                                     */
/*   dub -- double (IEEE 754 binary64) to receive an approximation */
/*          of num                                                 */
/*                                                                 */
/* Conventions vary for signalling NaNs, so all NaNs set dub to    */
/* NAN.                                                            */
/* --------------------------------------------------------------- */
void decToDouble(dec num, double *dub) {
  int exp;                         // decimal exponent

  if (isSpecial(num)) {            // a NaN or infinity
    if (isNaN(num)) *dub=NAN;
     else *dub=INFINITY;
    }
   else {                          // is finite
    *dub=(double)getCoeff(num);
    exp=getExp(num);
    if (exp!=0) {                  // not * 1
      if (exp>-308) *dub*=pow(10.0, exp);
       else { // pow(10, exp) would have fewer than 53 bits ..
        double h=pow(10.0, exp/2);
        *dub*=h;                   // .. so take two bites
        *dub*=h;                   // ..
        if (exp%2!=0) *dub*=0.1;   // one more /10 if exponent was odd
        }
      }
    }
  if (isSigned(num)) *dub=-*dub;   // [valid for -0, too]
  } // decToDouble

/* --------------------------------------------------------------- */
/* decToInt -- convert a decimal to int                            */
/*                                                                 */
/*   num -- decimal to convert                                     */
/*                                                                 */
/*   returns int result with DECBADINT if error                    */
/*                                                                 */
/* This converts the decimal number to an integer.  Fractional     */
/* parts of a finite number are discarded.  If the result is too   */
/* large to fit in a 32-bit int, or is a special value, then       */
/* DECBADINT is returned.                                          */
/* --------------------------------------------------------------- */
int decToInt(dec num) {
  int exp;                    // exponent
  int coe;                    // coefficient
  int need;                   // digits in answer

  if (isSpecial(num)) return DECBADINT;

  // is finite ...
  exp=getExp(num);            // extract ..
  coe=getCoeff(num);          // ..
  if (exp<0) {                            // reducing exponent
    if (exp<-9) return 0;
    coe/=powers[-exp];
    }
   else if (exp>0) {                      // positive exponent
    need=decGetDigits(num)+exp;
    if (need>10) return DECBADINT;        // too big
    if (need<10) coe=coe*powers[exp];     // will fit
     else {                               // need 10 digits; trickier
      coe*=powers[exp-1];                 // * 10 more needed
      if (coe>214748364) return DECBADINT;
      coe*=10;                            // complete
      }
    }
  // apply sign and return
  if (isSigned(num)) coe=-coe; // apply sign (cannot give DECBADINT)
  return coe;
  } // decToInt

/* --------------------------------------------------------------- */
/* decToString -- decimal to string                                */
/*                                                                 */
/*   num -- decimal to convert                                     */
/*   str -- string to receive converted decimal (char[17] +)       */
/*                                                                 */
/* Converts the decimal to a string, using the conventions as      */
/* recommended by IEEE 754 and specified by the General Decimal    */
/* Arithmetic specification.  For specials, the strings "NaN" and  */
/* "Infinity" are used.  The string is terminated by '\0'; up to   */
/* 17 characters (including the terminator) might be needed.       */
/* --------------------------------------------------------------- */
void decToString(dec num, char *str) {
  char *c=str;                // work
  int digits, exp, pre;       // ..
  int coeff;                  // ..

  // decPrint("dTS", num);

  // sign is allowed on all decimals
  if (isSigned(num)) {*c='-'; c++;}

  if (isSpecial(num)) {
    const char *out="Infinity";    // most likely
    if (isNaN(num)) {
      if (isSNaN(num)) {*c='s'; c++;}
      out="NaN";
      }
    for (; *out; out++, c++) *c=*out;
    *c='\0';                       // terminate
    return;
    }

  // finite number
  exp=getExp(num);
  digits=decGetDigits(num);
  coeff=getCoeff(num);

  // non-0 exponent -- assume plain form
  pre=digits+exp;                  // digits before '.'
  if (exp!=0) {
    if (exp<0 && pre>-6) exp=0;    // can use 'plain' form
     else {                        // will use exponential form
      exp=exp+digits-1;            // calculate adjusted E value
      pre=1;                       // request one digit before '.'
      } // need exponent

    if (pre<=0) {                  // 0.xxx or 0.000xxx form
      *c='0'; c++;
      *c='.'; c++;
      for (; pre<0; pre++, c++) *c='0'; // add any 0's after '.'
      pre=digits;                  // [to prevent another '.']
      }
    }

  // lay out the coefficient digits, inserting a '.' if required
  for (; digits>0; pre--, digits--, c++) {
    if (pre==0) {*c='.'; c++;}
    *c=coeff/powers[digits-1]+'0';
    coeff%=powers[digits-1];
    }

  // add the E part, if needed
  if (exp!=0) {
    int d;                         // work
    dec nume;                      // ..
    *c='E'; c++;
    if (exp>0) *c='+';             // positive
     else {*c='-'; exp=-exp;}      // negative
    c++;
    // lay out the exponent digits
    nume=decFromInt(exp);          // place ..
    d=decGetDigits(nume);          // .. to use counter
    for (; d>0; c++, d--) {
      *c=exp/powers[d-1]+'0';
      exp%=powers[d-1];
      }
    }

  *c='\0';                         // terminate
  return;
  } // decToString

/* --------------------------------------------------------------- */
/* streq -- caseless string compare                                */
/*                                                                 */
/*   char1 -- first string to compare                              */
/*   char2 -- second string                                        */
/*                                                                 */
/*   returns 1 if the strings caseless-compare equal, 0 otherwise  */
/*                                                                 */
/* If both strings are empty they are equal.                       */
/* --------------------------------------------------------------- */
static int streq(const char *char1, const char *char2) {
  for (;;char1++, char2++) {
    if (*char1==*char2) {if (*char1=='\0') return 1;}
     else if (toupper(*char1)!=toupper(*char2)) return 0;
    } // forever
  } // streq
