/* --------------------------------------------------------------- */
/* miniDec.h -- a decimal package for small systems                */
/* --------------------------------------------------------------- */
/*                                                                 */
/* Copyright (c) Mike Cowlishaw, 2010-2011.  All rights reserved.  */
/*                                                                 */
/* Permission to use, copy, modify, and/or distribute this         */
/* software for any purpose with or without fee is hereby granted, */
/* provided that the above copyright notice and this permission    */
/* notice appear in all copies, and that notice and the date of    */
/* any modifications be added to the software.                     */
/*                                                                 */
/* This software is provided "as is".  No warranties, whether      */
/* express, implied, or statutory, including, but not limited to,  */
/* implied warranties of merchantability and fitness for a         */
/* particular purpose apply to this software.  The author shall    */
/* not, in any circumstances, be liable for special, incidental,   */
/* or consequential damages, for any reason whatsoever.            */
/*                                                                 */
/* --------------------------------------------------------------- */
/* This package supports a simple decimal format, intended for the */
/* storage, management, and display of decimal data.  As such, it  */
/* provides conversions to and from ints, doubles, and strings in  */
/* various forms, along with decQuantize to 'shape' decimal data.  */
/* --------------------------------------------------------------- */
// NaNs currently have no payload, but this could be
// added later (probably only affects to/from string).
//
// Inspired by decNumber and decFloats, but all new code
//
// 2011.01.16 Rework to true BCD and conforming to Tollos code
//            conventions.
// 2011.01.20 Revise approach to exclude arithmetic for now and
//            provide small and fast conversions to and from double.
// 2011.01.23 Experiments showed that a 48-bit format would allow
//            significantly reduced call overhead on ARM, so revised
//            format (again)
// 2011.01.26 1.00 good

#ifndef MINIDEC_USED
#define MINIDEC_USED

#define MINIDECVERSION  "1.01"               /* package version */
#define MINIDECAUTHOR   "Mike Cowlishaw"     /* who to blame */

#include <stdint.h>
#include <limits.h>
#if !defined(MINIDECPRINT) | MINIDECPRINT==0 /* debug approach */
  #define MINIDECDEBUG debugf
#else
  #include <stdio.h>
  #define MINIDECDEBUG printf
#endif

/* A decimal number -- 48 bits */
typedef struct {
  uint32_t coef;    // coefficient 0..999999999 binary encoded
  int16_t  meta;    // signed unbiased exponent q [DECMINQ -> DECMAXQ]
                    // in top 12 bits; decimal sign and special flags
                    // in low 4 bits
  } dec;

// Sign and specials flags, held in lowest 4 bits of the meta field.
// for specials, the exponent and coefficient must be 0 (the
// coefficient could hold a NaN payload, later)
#define DECSIGN  0x08         // sign (1 = negative)
#define DECQNAN  0x04         // qNaN; ordering as per 754
#define DECSNAN  0x02         // sNaN
#define DECINF   0x01         // Infinity

// error indicator
#define DECBADINT 0x80000000  // 'Bad integer'

// Initializers for constants
#define DECINITZERO  {0, 0}
#define DECINITZEROM {0, DECSIGN}
#define DECINITQNAN  {0, DECQNAN}
#define DECINITSNAN  {0, DECSNAN}
#define DECINITINF   {0, DECINF}
#define DECINITINFM  {0, DECINF | DECSIGN}

/* Limits */
#define DECDIGITS          9            // maximum digits in coefficient
#define DECMAXCOE  999999999            // maximum coefficent as an int
#define DECMINCOE -999999999            // minimum signed coefficient
#define DECEMAX          999            // emax
#define DECEMIN         -998            // emin=1-emax
#define DECQMAX      DECEMAX            // maximum exponent q
#define DECQMIN   (DECEMIN-DECDIGITS+1) // minimum exponent q
#define DECEDIGITS         4            // maximum digits in exponent q

/* Constants */
extern const dec decZero;
extern const dec decZeroM;
extern const dec decNaN;
extern const dec decSNaN;
extern const dec decInf;
extern const dec decInfM;
extern const int powers[10];            // Powers of ten (0<=n<=9)

/* Debugging Macros */
#define decZSigned(d) ((d).meta&DECSIGN)
#define decZSpec(d) (               \
  ((d).meta&0x07)==0 ? ' ' :        \
  ((d).meta&DECSNAN) ? 'S' :        \
  ((d).meta&DECQNAN) ? 'Q' :        \
  ((d).meta&DECINF)  ? 'i' : '?')
#define decPrint(tag, d) {MINIDECDEBUG("%s: %c%09d x E%+d  %c\n",   \
                          (tag),                                    \
                          decZSigned(d) ? '-' : '+',                \
                          (d).coef,                                 \
                          (d).meta>>4,                              \
                          decZSpec(d))                              \
                          ;}

/* Routines */
extern dec  decFromDouble(double dub);
extern dec  decFromInt(int i);
extern dec  decFromInts(int coe, int exp);
extern dec  decFromString(const char *str);
extern int  decGetCoeff(dec num);
extern int  decGetDigits(dec num);
extern int  decGetExp(dec num);
extern int  decGetSign(dec num);
extern int  decIdentical(dec num1, dec num2);
extern int  decIsFinite(dec num);
extern int  decIsInf(dec num);
extern int  decIsNaN(dec num);
extern int  decIsNegative(dec num);
extern int  decIsPositive(dec num);
extern int  decIsQNaN(dec num);
extern int  decIsSigned(dec num);
extern int  decIsSNaN(dec num);
extern int  decIsSpecial(dec num);
extern int  decIsZero(dec num);
extern dec  decQuantize(dec a, dec b);
extern void decSetCoeff(dec *num, int coe);
extern void decSetExp(dec *num, int exp);
extern void decSetSign(dec *num, int sign);
extern void decToDouble(dec num, double *dub);
extern int  decToInt(dec num);
extern void decToString(dec num, char *str);

// ensure 'int' is 32-bit signed
#if (INT_MAX != 2147483647)
  #error miniDec assumes 'int' is 32 bits, and signed
#endif

#endif
