/* --------------------------------------------------------------- */
/* tollos.h -- core header file for Tollos                         */
/* --------------------------------------------------------------- */
/*                                                                 */
/* Copyright (c) Mike Cowlishaw, 2010-2011.  All rights reserved.  */
/*                                                                 */
/* Permission to use, copy, modify, and/or distribute this         */
/* software for any purpose with or without fee is hereby granted, */
/* provided that the above copyright notice and this permission    */
/* notice appear in all copies, and that notice and the date of    */
/* any modifications be added to the software.                     */
/*                                                                 */
/* This software is provided "as is".  No warranties, whether      */
/* express, implied, or statutory, including, but not limited to,  */
/* implied warranties of merchantability and fitness for a         */
/* particular purpose apply to this software.  The author shall    */
/* not, in any circumstances, be liable for special, incidental,   */
/* or consequential damages, for any reason whatsoever.            */
/*                                                                 */
/* --------------------------------------------------------------- */
/*                                                                 */
/* This includes:                                                  */
/*   Type and constant definitions                                 */
/*   Core API definitions                                          */
/*   Headers for other Tollos modules and utilities                */
/*                                                                 */
/* Device-level functions can include system I/O interfaces        */
/* in any Tollos header by defining TOLLOS_IO before including     */
/* this header.  Applications should not normally do this.         */
/*                                                                 */
/* --------------------------------------------------------------- */

#ifndef TOLLOS_USED
#define TOLLOS_USED

#define TOLLOS_VERSION   "tollos 1.22"       // Version [16 max.]
#define TOLLOS_AUTHOR    "Mike Cowlishaw"    // Who to blame

/* Standard includes */
#include <ctype.h>
#include <limits.h>
#include <stdlib.h>
#include <stdint.h>

/* Tollos type names -- for compatibility with various libraries */
typedef int8_t    flag;
typedef uint8_t   byte;
typedef uint8_t   uchar;
typedef int8_t    int8;
typedef uint8_t   uint8;
typedef int16_t   int16;
typedef uint16_t  uint16;
typedef int32_t   int32;
typedef uint32_t  uint32;
typedef uint32_t  uint;
typedef int64_t   int64;
typedef uint64_t  uint64;

// vector -- address of a function with no argument or return value
typedef void(*vector)(void);

// Basic Vector Table (must be a structure due to the mix of types)
typedef struct {
  void * v0;                  //  0: stack pointer
  vector v1;                  //  1: entry point to start
  vector v2;                  //  2: NMI
  vector v3;                  //  3: HardFault
  vector v4;                  //  4: MemManage
  vector v5;                  //  5: BusFault
  vector v6;                  //  6: UsageFault
  vector v7;                  //  7: (reserved)
  vector v8;                  //  8: (reserved)
  vector v9;                  //  9: (reserved)
  vector va;                  // 10: (reserved)
  vector vb;                  // 11: SVC
  vector vc;                  // 12: DebugMonitor
  vector vd;                  // 13: (reserved)
  vector ve;                  // 14: PendSVC
  vector vf;                  // 15: SysTick
  } basicvectors;

/* Timing constants [ms] */
#define TOLLOS_TICKTIME 40         // Default tick time for 25 Hz ticks
#define TOLLOS_MAXWATCHDOG 10000   // Maximum possible watchdog time

/* General named constants */

// Named port numbers
#define PORT0  0              // NXP names
#define PORT1  1
#define PORT2  2
#define PORT3  3
#define PORT4  4

#define PORTA  0              // STM names
#define PORTB  1
#define PORTC  2
#define PORTD  3
#define PORTE  4
#define PORTF  5
#define PORTG  6

// Channel numbers (for multiple I2C buses, etc.)
#define CHANNEL0 0
#define CHANNEL1 1
#define CHANNEL2 2

// Named pin numbers (pins in ports (or on boards) as numbers rather
// than masks; for the latter, see tollosBit.h)
#define PIN0    0
#define PIN1    1
#define PIN2    2
#define PIN3    3
#define PIN4    4
#define PIN5    5
#define PIN6    6
#define PIN7    7
#define PIN8    8
#define PIN9    9
#define PIN10  10
#define PIN11  11
#define PIN12  12
#define PIN13  13
#define PIN14  14
#define PIN15  15
#define PIN16  16
#define PIN17  17
#define PIN18  18
#define PIN19  19
#define PIN20  20
#define PIN21  21
#define PIN22  22
#define PIN23  23
#define PIN24  24
#define PIN25  25
#define PIN26  26
#define PIN27  27
#define PIN28  28
#define PIN29  29
#define PIN30  30
#define PIN31  31

// control character constants
#define ASCII_TAB 0x09
#define ASCII_LF  0x0a
#define ASCII_FF  0x0c
#define ASCII_CR  0x0d

/* Components of the Tollos core library */
#include "tollosBit.h"        // bits (GPIO)
#include "tollosDebug.h"      // debug serial port
#include "tollosDraw.h"       // display device(s)
#include "tollosFont.h"       // font descriptors
#include "tollosI2C.h"        // I2C bus devices
#include "tollosLED.h"        // LED(s)
#include "tollosSPI.h"        // SPI bus device(s)
#include "tollosUART.h"       // serial device
#include "tollosUtil.h"       // utilities and macros
#include "tollosVCom.h"       // USB Virtual COM port

/* Other generic interfaces in the Tollos library */
// (these may be moved to a separate library or libraries later)
#include "tollosAcc.h"        // accelerometers
#include "tollosGyro.h"       // gyroscopes

/* Non-Tollos module */
#define   MINIDEC 1           // indicate available
#define   MINIDECPRINT 0      // don't use printf
#include "miniDec.h"

/* --------------------------------------------------------------- */
/* Tollos core API                                                 */
/* --------------------------------------------------------------- */

/* Constants for sysSet items */
#define SYS_TICKMAX      101  // maximum tick() call time, ms
#define SYS_TICKTIME     102  // tick interval, ms
#define SYS_WATCHDOG     103  // watchdog interval, ms

/* Constants for sysQuery items */
#define SYS_CLOCK        201  // clock speed, Hz
#define SYS_RESET        202  // 0=normal, 1=watchdog or brownout
#define SYS_TICKLAST     203  // time for last tick() call, ms
#define SYS_STACKSIZE    204  // total RAM available for stack
#define SYS_STACKUSED    205  // low water mark; maximum stack used
#define SYS_TICKFLAG     206  // toggles 0<-->1 every 1 ms
// (items valid for sysSet are always also valid for sysQuery)

extern void sysPause(uint ms);               // wait awhile
extern int  sysQuery(uint item);             // query value of an item
extern int  sysSet(uint item, int value);    // set an item
extern void sysStartup(void);                // soft reset (startup)
extern int  sysVector(uint index, vector addr, uint level); // set interrupt

/* --------------------------------------------------------------- */
/* Application/user functions                                      */
/*                                                                 */
/* init -- is called once, after C and processor initialization    */
/* tick -- is called every systick after init completes            */
/*                                                                 */
/* They should normally return 0; see sysStartup commentary for    */
/* more details.  Either can be omitted.                           */
/* --------------------------------------------------------------- */
extern int init(void) __attribute__ ((weak));
extern int tick(void) __attribute__ ((weak));

/* --------------------------------------------------------------- */
/* Tollos system functions                                         */
/* --------------------------------------------------------------- */
#ifdef TOLLOS_IO
  #include "tollosSystem.h"   // System-dependent interface
#endif

/* --------------------------------------------------------------- */
/* Assertions                                                      */
/* --------------------------------------------------------------- */
#if NULL != 0
  #error NULL has unexpected value
#endif

// ensure 'int' is 32-bit signed
#if (INT_MAX != 2147483647)
  #error Tollos assumes 'int' is 32 bits
#endif
#endif
