/* --------------------------------------------------------------- */
/* tollos.ld -- Generic linker script for ARM-based chips          */
/* --------------------------------------------------------------- */
/*                                                                 */
/* Copyright (c) Mike Cowlishaw, 2010-2011.  All rights reserved.  */
/*                                                                 */
/* Permission to use, copy, modify, and/or distribute this         */
/* software for any purpose with or without fee is hereby granted, */
/* provided that the above copyright notice and this permission    */
/* notice appear in all copies, and that notice and the date of    */
/* any modifications be added to the software.                     */
/*                                                                 */
/* This software is provided "as is".  No warranties, whether      */
/* express, implied, or statutory, including, but not limited to,  */
/* implied warranties of merchantability and fitness for a         */
/* particular purpose apply to this software.  The author shall    */
/* not, in any circumstances, be liable for special, incidental,   */
/* or consequential damages, for any reason whatsoever.            */
/*                                                                 */
/* --------------------------------------------------------------- */
/*                                                                 */
/* This script is included by chip-specific .ld scripts, which     */
/* set the following symbols before including this:                */
/*                                                                 */
/*   chipRAMbase      -- RAM origin                                */
/*   chipRAMsize      -- RAM size                                  */
/*   chipROMbase      -- ROM (flash) origin                        */
/*   chipROMsize      -- ROM size [e.g., 512K]                     */
/*   chipSRAM0base    -- SRAM0 origin                              */
/*   chipSRAM0size    -- SRAM0 size                                */
/*   chipSRAM1base    -- SRAM1 origin                              */
/*   chipSRAM1size    -- SRAM1 size                                */
/*                                                                 */
/*   chipCAN          -- 0=none, n=number available                */
/*   chipDAC          -- 0=none, n=number available                */
/*   chipEthernet     -- 0=none, n=number available                */
/*   chipI2C          -- 0=none, n=number available                */
/*   chipI2S          -- 0=none, n=number available                */
/*   chipUSB          -- 0=none, 1=Device, 3=Device/Host/OTG       */
/*                                                                 */
/*   chipMaxClock     -- Hz [e.g., 100000000]                      */
/*   chipVectors      -- total number of interrupt vectors         */
/*                       (including the 16 base vectors)           */
/*                                                                 */
/* The chip-specific scripts are in turn included by board-        */
/* specific scripts, which set:                                    */
/*                                                                 */
/*   chipCrystal     -- crystal/oscillator frequency, Hz           */
/*   chipOscillator  -- [optional] 1=oscillator, 0=crystal         */
/*                                                                 */
/* --------------------------------------------------------------- */
/*                                                                 */
/* The following symbols are set by this script and are also       */
/* visible to C code if required; use the & address operator:      */
/*                                                                 */
/* They should apply to all ARM Cortex M3 chips:                   */
/*                                                                 */
/*   chipVectorSource -- start of the basic Vector table in ROM    */
/*   chipVectorSink   -- end of ditto +1                           */
/*   chipVectorStart  -- start of the full Vector table in RAM     */
/*   chipVectorEnd    -- end of ditto +1                           */
/*                                                                 */
/*   chipDataSource   -- start of the static initialization in ROM */
/*   chipDataStart    -- start of the initialized statics in RAM   */
/*   chipDataEnd      -- end of ditto +1                           */
/*                                                                 */
/*   chipZeroStart    -- start of zeroed statics in RAM            */
/*   chipZeroEnd      -- end of ditto +1                           */
/*                                                                 */
/*   chipRAMtop       -- highest used address in RAM, +1           */
/*   chipRAMend       -- end of RAM, +1                            */
/*   chipROMend       -- end of ROM, +1                            */
/*                                                                 */
/*   chipVectorSpace  -- size of the RAM vector table [bytes]      */
/* --------------------------------------------------------------- */

chipRAMend      = chipRAMbase + chipRAMsize;
chipROMend      = chipROMbase + chipROMsize;
chipVectorSpace = chipVectors * ABSOLUTE(4);

/* Notes:
 *
 * Into ROM goes
 *   -- interrupt vector table (core vectors only)
 *   -- code (.text etc.)
 *   -- read-only constant data (.rodata etc.)
 *   -- initialization values for explicitly-initialized statics
 * In RAM, addresses are set up for
 *   -- writable interrupt vector table (all vectors)
 *   -- initialized statics (copied from ROM at startup)
 *   -- zeroed statics (cleared at startup)
 *   -- remaining space, all stack
 *
 * There is no MEMORY definition here as it cannot use symbols;
 *   plausible stack space checking can be done at startup; stack
 *   usage can be inspected dynamically.
 *
 * The KEEP() around *(vectors) forces that section to be included,
 * so that if -gc-sections is used the vectors are not discarded.
 * The vectors (and startup) must not be in a .a file if they are
 * to be included in all cases.
 *
 * In the following sections, data have:
 *  -- a load address (by default the section preceding)
 *  -- a relocation address (where it will be addressed at run time)
 *
 */
SECTIONS {

    .text chipROMbase : {     /* starting at the base of ROM */
      chipVectorSource = .;
      KEEP(*(vectors))        /* vector table; must be at ROM base */
      chipVectorSink = .;

      *(.text .text.*)        /* program code */
      *(.rodata .rodata*)     /* read-only (const) static data */

      . = ALIGN(4);
      chipTextEnd = .;        /* = start of ROM copy of remaining data */
      }

    /* The following have a relocation address in RAM but are also */
    /* loaded into ROM if data are provided                        */

    .ramvectors chipRAMbase : AT (chipTextEnd) {
      chipVectorStart = .;
      . += chipVectorSpace;   /* just leave space; no related section */
      . = ALIGN(4);
      chipVectorEnd = .;
      }

    .data : {
      chipDataSource = LOADADDR(.data);
      chipDataStart = .;
      *(.data .data.*)        /* initialized static data (non-const) */
      . = ALIGN(4);
      chipDataEnd = .;
      }

    .bss : {
      chipZeroStart = .;
      *(.bss)                 /* zero-filled static data (non-const) */
      . = ALIGN(4);
      chipZeroEnd = .;
      chipRAMtop = .;
      }

    }

