/* --------------------------------------------------------------- */
/* tollosAcc.c -- accelerometer functions for Tollos               */
/* --------------------------------------------------------------- */
/*                                                                 */
/* Copyright (c) Mike Cowlishaw, 2010-2011.  All rights reserved.  */
/*                                                                 */
/* Permission to use, copy, modify, and/or distribute this         */
/* software for any purpose with or without fee is hereby granted, */
/* provided that the above copyright notice and this permission    */
/* notice appear in all copies, and that notice and the date of    */
/* any modifications be added to the software.                     */
/*                                                                 */
/* This software is provided "as is".  No warranties, whether      */
/* express, implied, or statutory, including, but not limited to,  */
/* implied warranties of merchantability and fitness for a         */
/* particular purpose apply to this software.  The author shall    */
/* not, in any circumstances, be liable for special, incidental,   */
/* or consequential damages, for any reason whatsoever.            */
/*                                                                 */
/* --------------------------------------------------------------- */
/* These functions provide basic accelerometer functions for       */
/* Tollos, intended to be device-independent (although at the      */
/* moment an I2C connection is assumed).                           */
/*                                                                 */
/* To use, the device must be initialized using accInit.  This     */
/* indicates which I2C channel the device is on and which port/pin */
/* is used for DataReady interrupt.  By default the accelerometer  */
/* range will be at least 2g; this may be changed with accRange.   */
/*                                                                 */
/* The latest accelerometer reading can be read safely using       */
/* accGetXYZ, which returns the three axes in milli-g.  The axes   */
/* directions are as defined by the device manufacturer.           */
/*                                                                 */
/* The default update rate is selected by the device driver code   */
/* and will be at least 25Hz (that is, the data will be no more    */
/* than 40ms old).  Update is not necessarily synchronous with the */
/* Tollos tick rate.   The rate can be changed with accRate.       */
/* --------------------------------------------------------------- */

#define  TOLLOS_IO            // allow use of I/O functions
#include "tollos.h"

// vector of readings from Accelerometer
typedef struct {
  int x;                      // X reading [mg]
  int y;                      // Y reading [mg]
  int z;                      // Z reading [mg]
  } accData;

// Double-buffer of readings; this is used to avoid race conditions
// between the interrupt handler and the accGetXYZ function as
// follows:
//
//   1. On entry to the Get function, accLast is flipped 0<->1.
//      (It is assumed that the interrupt handler cannot be running
//      at the moment accLast is actually written, but even if it
//      were, the buffer chosen would be arbitrary but would be
//      wholly written before the Get function regained control.)
//
//   2. The Get function tests accGoodData[accLast]; if 0, no good
//      data is available and this is reported to the caller.
//
//   3. If an accBuff does have good data this is set to the caller;
//      then the accGoodData flag is cleared.
//
//   4. In the interrupt handler, data are read from the device and
//      written to the non-accLast buffer, then accGoodData for that
//      buffer is set to 1 (this may happen many times, overwriting
//      the data, until there is a read).
//
// Hence: the two levels of code always write to and read from
// different buffers; access to those buffers is controlled by
// accLast alone, which is only switched in the lower priority
// task (when it is not accessing the buffer).
//
static accData accBuff[2];         // double buffer
static flag    accGoodData[2];     // 1 if the corresponding accBufff
                                   // has good data
static uint    accLast=0;          // which buffer was last read


// local function
static void accHandler(void); // interrupt handler

/* --------------------------------------------------------------- */
/* accGetXYZ -- read latest x, y, and z accelerations              */
/*                                                                 */
/*   x, y, z -- receive accelerations [mg: 1000 = 1.000 g], or all */
/*     zero if error                                               */
/*                                                                 */
/*   returns 0 if data good, -1 if error (no good data since last  */
/*     read, or device not initialized)                            */
/*                                                                 */
/* Results are returned scaled to milli-gs (1000 = 1g), without    */
/* any calibration adjustments.                                    */
/*                                                                 */
/* This must not be called from an interrupt level with higher     */
/* priority than the accelerometer interrupt.                      */
/* --------------------------------------------------------------- */
int accGetXYZ(int *x, int *y, int*z) {
  // see above for the handshaking protocol used here
  accLast=!accLast;
  if (!accGoodData[accLast]) {     // no good data
    *x=0; *y=0; *z=0;
    return -1;
    }

  // return data from the 'good' buffer
  *x=accBuff[accLast].x;
  *y=accBuff[accLast].y;
  *z=accBuff[accLast].z;

  accGoodData[accLast]=0;          // indicate data used

  return 0;                        // is good
  } // accGetXYZ

/* --------------------------------------------------------------- */
/* accHandler -- interrupt handler, called when Data Ready         */
/*                                                                 */
/* This reads the data into the 'active' buffer, then when and if  */
/* good sets accGoodData=1 for that buffer.                        */
/* --------------------------------------------------------------- */
static void accHandler(void) {
  int rc;                          // work
  int read=!accLast;               // buffer to read into

  rc=accGetXYZIO(&accBuff[read].x, &accBuff[read].y, &accBuff[read].z);
  // interrupt is cleared by the read
  if (rc<0) return;                // bad data

  accGoodData[read]=1;             // have goodness
  } // accHandler

/* --------------------------------------------------------------- */
/* accInit -- configure and initialize accelerometer               */
/*                                                                 */
/*   channel -- I2C channel on which device is to be found         */
/*   drport  -- port for the Data Ready interrupt pin              */
/*   drpin   -- pin number (not bit mask) for the Data Ready       */
/*              interrupt pin                                      */
/*                                                                 */
/*   returns 0 if initialized OK, negative if error                */
/*                                                                 */
/* This sets up the pin and interrupt handler for the Data Ready   */
/* interrupt from the device then calls the device driver          */
/* initialization function to start the device.                    */
/*                                                                 */
/* The accInitIO function will set:                                */
/* * an update rate of at least 25 Hz (but not necessarily higher  */
/* than that, to save power)                                       */
/* * a range of at least 2g (call accRange if a larger range is    */
/* needed)                                                         */
/* * an interrupt on Data Ready.                                   */
/*                                                                 */
/* Note that it can be 40ms or more before the first valid data    */
/* are available, so an accGetXYX call before that time may report */
/* no data available, as might a call that reads the device less   */
/* than 25ms after the last read.                                  */
/* --------------------------------------------------------------- */
int accInit(uint channel, uint drport, uint drpin) {
  int rc;                     // work

  accLast=0;                  // must be 0 or 1
  accGoodData[0]=0;           // no good data yet
  accGoodData[1]=0;           // ..

  // set up interrupt pin from number (bitFuns use a mask)
  rc=bitEnable(drport, 1<<drpin, BIT_RISE, accHandler);

  if (rc==0) rc=accInitIO(channel);
  return rc;
  } // accInit

/* --------------------------------------------------------------- */
/* accRange -- change range of accelerometer                       */
/*                                                                 */
/*   range -- range requested in milli-g [1000 = 1 g]              */
/*                                                                 */
/*   returns 0 if range changed OK, -1 if error (range too high)   */
/*                                                                 */
/* The accelerometer will be set to a smallest available range     */
/* that is >= the requested range, to maximize the accuracy of     */
/* returned results.                                               */
/* --------------------------------------------------------------- */
int accRange(uint range) {
  return accRangeIO(range);   // pass on and back
  } // accRange

/* --------------------------------------------------------------- */
/* accRate -- change update rate of accelerometer                  */
/*                                                                 */
/*   rate -- rate requested [Hz, 1..200]                           */
/*                                                                 */
/*   returns 0 if rate changed OK, -1 if error (rate zero or       */
/*     too high)                                                   */
/*                                                                 */
/* The accelerometer will be set to a slowest available rate       */
/* that is >= the requested rate, to minimize resource usage.      */
/*                                                                 */
/* To ensure working with 100kHz I2C, the maximum rate is limited  */
/* to 200Hz -- but note that this may use significant processor    */
/* bandwidth.                                                      */
/* --------------------------------------------------------------- */
int accRate(uint rate) {
  return accRateIO(rate);     // pass on and back
  } // accRate

/* --------------------------------------------------------------- */
/* accStop -- stop and power-down accelerometer                    */
/*                                                                 */
/*   returns 0 if accelerometer stopped OK, negative if an error   */
/*     or it was not running.                                      */
/* --------------------------------------------------------------- */
int accStop(void) {
  int rc=accStopIO();         // stop the device
  accGoodData[0]=0;           // invalidate any data
  accGoodData[1]=0;           // ..
  return rc;
  } // accStop

