/* --------------------------------------------------------------- */
/* tollosDebug.c -- debug functions for Tollos                     */
/* --------------------------------------------------------------- */
/*                                                                 */
/* Copyright (c) Mike Cowlishaw, 2010-2011.  All rights reserved.  */
/*                                                                 */
/* Permission to use, copy, modify, and/or distribute this         */
/* software for any purpose with or without fee is hereby granted, */
/* provided that the above copyright notice and this permission    */
/* notice appear in all copies, and that notice and the date of    */
/* any modifications be added to the software.                     */
/*                                                                 */
/* This software is provided "as is".  No warranties, whether      */
/* express, implied, or statutory, including, but not limited to,  */
/* implied warranties of merchantability and fitness for a         */
/* particular purpose apply to this software.  The author shall    */
/* not, in any circumstances, be liable for special, incidental,   */
/* or consequential damages, for any reason whatsoever.            */
/*                                                                 */
/* --------------------------------------------------------------- */
/* The debug device uses a serial channel to communicate with a    */
/* terminal device (usually a termulator emulator application on   */
/* a PC).                                                          */
/*                                                                 */
/* Using any debug function automatically initializes the debug    */
/* communication device.                                           */
/* The default baud rate for the device is TOLLOS_DEBUG_BAUD,      */
/* normally 115200 baud.                                           */
/* --------------------------------------------------------------- */

#define  TOLLOS_IO            // allow use of I/O functions
#include "tollosDebug.h"
#include <stdarg.h>
#include <string.h>

static flag initialized=0;    // 1 if initialized

/* --------------------------------------------------------------- */
/* debugBaud -- set debug baud rate                                */
/*                                                                 */
/*   baud -- baud rate to use for debug                            */
/*                                                                 */
/* It is recommend that a 'standard' baud rate be used, for        */
/* example, 9600 or 115200.                                        */
/* At 9600 baud about 38 characters can be sent in 40ms (one       */
/* default tick) at 115200, 460 characters can be sent.            */
/* --------------------------------------------------------------- */
void debugBaud(uint baud) {   // initialize
  debugInitIO(baud);
  sysPause(0);                // dummy link
  initialized=1;
  } // debugBaud

/* --------------------------------------------------------------- */
/* debugf -- a mini printf to send a string                        */
/*                                                                 */
/*   format -- format and data string                              */
/*   ...    -- other arguments as required                         */
/*                                                                 */
/*   returns the number (count) of characters sent                 */
/*                                                                 */
/* See the tollosUtil ""strf"" function for a description of       */
/* supported format strings (a subset of the C printf rules).      */
/* --------------------------------------------------------------- */
int debugf(const char *format, ...) {
  va_list args;          // argument list
  int count=0;           // chars put counter

  if (!initialized) debugBaud(TOLLOS_DEBUG_BAUD);

  va_start(args, format);

  // carry out formatting, sending generated characters to debugPutcIO
  count=strf(debugPutcIO, format, args);

  va_end(args);          // done with argument list
  return count;
  } // debugf

/* --------------------------------------------------------------- */
/* debugGet -- get character from host                             */
/*                                                                 */
/*   returns a character read, or -1 if none pending               */
/*                                                                 */
/* If a character was read the value returned will always be       */
/* non-negative (hence all values >=0 might be possible).          */
/*                                                                 */
/* Note that some keyboard keys may generate more than one         */
/* character: an escape followed by some code, for example.        */
/* --------------------------------------------------------------- */
int debugGet(void) {          // get character
  if (!initialized) debugBaud(TOLLOS_DEBUG_BAUD);
  return debugGetcIO();
  } // debugGet

