/* --------------------------------------------------------------- */
/* tollosDraw.c -- draw (display) device functions                 */
/* --------------------------------------------------------------- */
/*                                                                 */
/* Copyright (c) Mike Cowlishaw, 2010-2011.  All rights reserved.  */
/*                                                                 */
/* Permission to use, copy, modify, and/or distribute this         */
/* software for any purpose with or without fee is hereby granted, */
/* provided that the above copyright notice and this permission    */
/* notice appear in all copies, and that notice and the date of    */
/* any modifications be added to the software.                     */
/*                                                                 */
/* This software is provided "as is".  No warranties, whether      */
/* express, implied, or statutory, including, but not limited to,  */
/* implied warranties of merchantability and fitness for a         */
/* particular purpose apply to this software.  The author shall    */
/* not, in any circumstances, be liable for special, incidental,   */
/* or consequential damages, for any reason whatsoever.            */
/*                                                                 */
/* --------------------------------------------------------------- */
/* The draw functions provide access to display device(s) (LCDs,   */
/* etc.), including text and graphics functions.                   */
/*                                                                 */
/* Display takes three steps:                                      */
/* # Initialize the display device using drawInit                  */
/* # Build up an image in the screen buffer (maintained by Tollos) */
/* by setting pels or using drawing functions such as drawLine     */
/* or text functions such as drawf (draw formatted string)         */
/* # Write the screen buffer to the device with drawWrite.         */
/*                                                                 */
/* The model here is that supported display devices have a         */
/* rectangular screen with a fixed number of pels (pixels)         */
/* addressed in X and Y, where the origin of both is zero, and     */
/* the pel addressed by X=0 and Y=0 is at the bottom left of the   */
/* screen.                                                         */
/*                                                                 */
/* The number of pels in X and Y (and other characteristics) are   */
/* reported in a drawConfig structure when the display device is   */
/* initialized.                                                    */
/*                                                                 */
/* The model for pels is that each has a value in the range 0..255 */
/* in red, green, and blue (where 0 is black and 255 is 100% of    */
/* each primary).  A given device (screen) may not support 8-bit   */
/* colours or even greyscale, so the nearest available value will  */
/* be used.  The chosen value can be determined by reading back    */
/* a pel.  The following drawRGB constants are provided; they are  */
/* initialized to the 16 IBM PC CGA colours: BLACK CGA_DARKBLUE    */
/* CGA_DARKGREEN CGA_DARKCYAN CGA_DARKRED CGA_DARKPINK CGA_BROWN   */
/* PALEGREY DARKGREY CGA_BLUE CGA_GREEN CGA_CYAN CGA_RED CGA_PINK  */
/* CGA_YELLOW WHITE.                                               */
/*                                                                 */
/* Tollos maintains state in addition to the screen buffer to      */
/* simplify use of the screen; this includes (defaults shown in    */
/* brackets):                                                      */
/* * current pen colour (used for drawing lines and text, etc.)    */
/* [black]                                                         */
/* * current canvas (background) color, used for clearing the      */
/* screen, a rectangle, or a pel and optionally used when drawing  */
/* text [white]                                                    */
/* * X position for next drawing operation [0]                     */
/* * Y position for next drawing operation [0]                     */
/* * The DRAW_BOUNDS flag [0].                                     */
/*                                                                 */
/* While drawing, errors include parameters out of range, such as  */
/* X or Y values outside the range of a 16-bit signed integer      */
/* (-32768..32767).  To report these, draw functions that receive  */
/* bad parameters set the DRAW_BOUNDS flag rather than returning a */
/* negative return code.  This allows a check for errors to be     */
/* made at convenient intervals rather than at after every         */
/* function call.  DRAW_BOUNDS is cleared when the screen is       */
/* cleared or written, as described in the relevant functions.     */
/*                                                                 */
/* Text processing:                                                */
/*                                                                 */
/* Tollos maintains additional state for text characters; this     */
/* includes (defaults shown in brackets):                          */
/* * current font; fonts are bitmap fonts held in fontInfo         */
/* structures [tollosFont5x7]                                      */
/* * X position for next character [0]                             */
/* * Y position for next character (baseline pel) [top of screen,  */
/* adjusted for font height and descent of first character]        */
/* * Whether to write pen, canvas, both, or neither kind of pel    */
/* from the character; the last of these can be used to measure    */
/* the width of characters, formatted strings, etc.                */
/*                                                                 */
/* When pen or canvas pels are written for a character the X and Y */
/* positions are first wrapped to the next row (or to the top of   */
/* the screen) if there is insufficient space for the character    */
/* and the character is then written.  In all cases the X position */
/* is then incremented by the width of the character, +1 optional  */
/* spacing.                                                        */
/* The leading (gap) between character rows is 1 pel (optional).   */
/* See the drawPutc function for more detail on text drawing.      */
/* --------------------------------------------------------------- */

#define  TOLLOS_IO                 // allow use of I/O functions
#include "tollos.h"
#include "tollosFont5x7.h"         // base system font

// screen configuration (initialized by the device glue, via drawInit)
static drawConfig screen;

// current tollosDraw state (initialized by drawInit)
static drawState state;            // queryable state
// the state structure is static, but as it is a unit a save/restore
// interface could be added later to permit control of multiple
// draw contexts/screens

// state lookaside
static int penvalue;               // current foreground pel value
static int canvalue;               // current background pel value

// constants for all to use -- the 16 CGA colours
// only the greys are 'pure' colours, so the others have a prefix
const drawRGB BLACK        ={0x00, 0x00, 0x00, 0x00}; //  0
const drawRGB CGA_DARKBLUE ={0x00, 0x00, 0x00, 0xaa}; //  1
const drawRGB CGA_DARKGREEN={0x00, 0x00, 0xaa, 0x00}; //  2
const drawRGB CGA_DARKCYAN ={0x00, 0x00, 0xaa, 0xaa}; //  3
const drawRGB CGA_DARKRED  ={0x00, 0xaa, 0x00, 0x00}; //  4
const drawRGB CGA_DARKPINK ={0x00, 0xaa, 0x00, 0xaa}; //  5
const drawRGB CGA_BROWN    ={0x00, 0xaa, 0x55, 0x00}; //  6
const drawRGB PALEGREY     ={0x00, 0xaa, 0xaa, 0xaa}; //  7
const drawRGB DARKGREY     ={0x00, 0x55, 0x55, 0x55}; //  8
const drawRGB CGA_BLUE     ={0x00, 0x55, 0x55, 0xff}; //  9
const drawRGB CGA_GREEN    ={0x00, 0x55, 0xff, 0x55}; // 10
const drawRGB CGA_CYAN     ={0x00, 0x55, 0xff, 0xff}; // 11
const drawRGB CGA_RED      ={0x00, 0xff, 0x55, 0x55}; // 12
const drawRGB CGA_PINK     ={0x00, 0xff, 0x55, 0xff}; // 13
const drawRGB CGA_YELLOW   ={0x00, 0xff, 0xff, 0x55}; // 14
const drawRGB WHITE        ={0x00, 0xff, 0xff, 0xff}; // 15

// local text defines and function
#define textTab 4                  // chars/tab [make variable later?]
static void linefeed(int);

/* --------------------------------------------------------------- */
/* drawBox -- draw a box with corner at current position           */
/*                                                                 */
/*   x    -- X coordinate of opposite corner                       */
/*   y    -- Y coordinate to opposite corner                       */
/*   fill -- 1 to fill the box with the canvas colour; 0 to leave  */
/*           the inside of the box unchanged                       */
/*                                                                 */
/* This draws a line box using the current pen colour.  The box is */
/* drawn such that the current position forms one corner and x, y  */
/* sets the opposite corner.  The pels within the line box will    */
/* be filled with the canvase colour if fill=1.                    */
/*                                                                 */
/* The current X and Y are not changed.                            */
/*                                                                 */
/* If x or y would not fit in 16 bits, DRAW_BOUNDS is set to 1 and */
/* no box (or fill) is drawn.                                      */
/* --------------------------------------------------------------- */
void drawBox(int x, int y, flag fill) {
  int x1, y1;                 // start point


  if (x!=(int16)x || y!=(int16)y) {
    state.drawBounds=1;
    return;
    }

  x1=state.drawX;             // save
  y1=state.drawY;             // ..

  if (fill) {
    int xlo=min(x, x1)+1;     // set X fill bounds
    int xhi=max(x, x1)-1;
    if (xhi>=xlo) {           // if horizontal pels to do
      int ylo=min(y, y1)+1;   // set Y fill bounds
      int yhi=max(y, y1)-1;
      int pen=penvalue;       // save current pen
      penvalue=canvalue;      // fill in canvas colour
      for (; ylo<=yhi; ylo++) {
        drawXY(xlo, ylo);
        drawLine(xhi, ylo);
        }
      penvalue=pen;           // restore
      drawXY(x1, y1);         // ..
      } // have x
    } // fill

  // draw the box surround
  drawLine(x1, y);
  drawLine(x,  y);
  drawLine(x,  y1);
  drawLine(x1, y1);           // (leaves current as x1, y1)
  } // drawBox

/* --------------------------------------------------------------- */
/* drawCanvas -- set the canvas (background) colour                */
/*                                                                 */
/*   rgb -- colour value [0..255 in each]                          */
/*                                                                 */
/* This sets the Canvas RGB value.                                 */
/* No error is possible.                                           */
/* --------------------------------------------------------------- */
void drawCanvas(drawRGB rgb) {
  // save value in lookaside and RGB in device context
  canvalue=drawFromRGBIO(rgb);     // convert to device value
  state.drawCanvas=rgb;
  } // drawCanvas

/* --------------------------------------------------------------- */
/* drawClear -- clear all screen pels to the canvas colour         */
/*                                                                 */
/* This clears the screen buffer and resets the text X and Y       */
/* position so the next character written will appear at top left. */
/* The DRAW_BOUNDS flag is cleared.                                */
/*                                                                 */
/* No other state is affected.                                     */
/* No error is possible.                                           */
/* --------------------------------------------------------------- */
void drawClear(void) {
  drawClearIO(canvalue);           // effect the clear
  state.textX=screen.width;        // current position (see drawInit)
  state.textY=0;                   // ..
  state.drawBounds=0;              // clear error flag
  } // drawClear

/* --------------------------------------------------------------- */
/* drawf -- draw a text string with formatting                     */
/*                                                                 */
/*   format -- format and data string                              */
/*   ...    -- other arguments as required                         */
/*                                                                 */
/*   returns the number (count) of characters sent                 */
/*                                                                 */
/* See the tollosUtil ""strf"" function for a description of       */
/* supported format strings (a subset of the C printf rules).      */
/* --------------------------------------------------------------- */
int drawf(const char *format, ...) {
  va_list args;          // argument list
  int count=0;           // chars put counter

  va_start(args, format);

  // carry out formatting, sending generated characters to drawPutc
  count=strf(drawPutc, format, args);

  va_end(args);          // done with argument list
  return count;
  } // drawf

/* --------------------------------------------------------------- */
/* drawFill -- fill a simple polygon                               */
/*                                                                 */
/*   x -- X coordinate to fill from                                */
/*   y -- Y coordinate to fill from                                */
/*                                                                 */
/* This fills an area bounded by pixels in the pen colour with     */
/* pixels in the canvas colour.  x and y give the start point of   */
/* the fill; if that point is in the pen colour no fill is carried */
/* out.                                                            */
/*                                                                 */
/* In order to limit the time and memory requirements of this      */
/* function, only simple shapes (in general, convex polygons) will */
/* be completely filled; the fill may be incomplete if there are   */
/* concave vertices (line junctions at an angle that points into   */
/* the fill area) or if there are 'islands' of pen colour pels     */
/* within the area.                                                */
/*                                                                 */
/* The fill area is assumed to be four-connected; that is, the     */
/* fill will stop on reaching a pen-coloured pel horizontally or   */
/* vertically -- it will not pass between pels that touch          */
/* diagonally (it can therefore be used to fill ellipses or        */
/* circles drawn by drawOval, etc.).                               */
/*                                                                 */
/* If x or y is offscreen, DRAW_BOUNDS is set to 1 and no fill is  */
/* attempted.                                                      */
/* --------------------------------------------------------------- */
void drawFill(int x, int y) {
  int look;              // current state: 0=looking both up and
                         // down, +1=looking/moving up, -1=down
  int scan;              // scan direction in x (-1 or +1)
  int upx, upy;          // pending up pel
  int dox, doy;          // pending down pel
  int sx, sy;            // start/save point

  if ((uint)x>=screen.width || (uint)y>=screen.height) {
    state.drawBounds=1;
    return;
    }
  if (drawGetPelIO(x, y)==penvalue) return;  // non-starter

  look=0;                // at start row; heads up and down
  upx=-1; upy=0;         // -1 to indicate no pending
  dox=-1; doy=0;         // .. [Ys set just to placate compiler]

  for (;;) {             // up/down loop
    for (scan=-1; scan<=1; scan+=2) { // scan left then right
      if (scan==-1) {sx=x; sy=y;}     // save start point
       else {x=sx+1; y=sy;}           // continue to the right

      // fill left or right, looking up & down as required
      for (; (uint)x < screen.width; x+=scan) {
        if (drawGetPelIO(x, y)==penvalue) break;  // reached bound
        drawSetPelIO(x, y, canvalue);             // fill current
        // if need to look and no pending yet, check for fillable
        if (look>=0 && upx<0 && y+1<(int)screen.height
         && drawGetPelIO(x, y+1)!=penvalue) {upx=x; upy=y+1;} // pending up
        if (look<=0 && dox<0 && y>0
         && drawGetPelIO(x, y-1)!=penvalue) {dox=x; doy=y-1;} // pending down
        }
      } // scan

    // finished both scans of a row
    if (upx>=0) {             // something above to do
      x=upx; y=upy;           // start here
      upx=-1;                 // none pending yet
      look=+1;                // need to look up only
      }
     else if (dox>=0) {       // something below to do
      x=dox; y=doy;           // start here
      dox=-1;                 // none pending yet
      look=-1;                // need to look down only
      }
     else break;              // nothing pending -- all done
    } // up/down
  } // drawFill

/* --------------------------------------------------------------- */
/* drawGetFont -- get the current font                             */
/*                                                                 */
/*   font -- set to the pointer to the current font                */
/*                                                                 */
/* No error is possible.                                           */
/* --------------------------------------------------------------- */
void drawGetFont(const fontInfo **font) {
  *font=state.textFont;
  } // drawGetFont

/* --------------------------------------------------------------- */
/* drawGetPel -- get the colour of a pel                           */
/*                                                                 */
/*   x   -- the X coordinate of the pel                            */
/*   y   -- the Y coordinate of the pel                            */
/*   rgb -- returned colour value [0..255 in each]                 */
/*                                                                 */
/* This returns the RGB values of the pel at x, y.                 */
/* If x or y is offscreen, 0 is returned for all three values.     */
/* If x or y would not fit in 16 bits, DRAW_BOUNDS is set to 1.    */
/* --------------------------------------------------------------- */
void drawGetPel(int x, int y, drawRGB *rgb) {
  if (x!=(int16)x || y!=(int16)y) state.drawBounds=1;
  drawToRGBIO(drawGetPelIO(x, y), rgb); // get pel, to RGB
  } // drawGetPel

/* --------------------------------------------------------------- */
/* drawInit -- initialize display and discover its configuration   */
/*                                                                 */
/*   config -- configuration structure (always set)                */
/*                                                                 */
/*   returns 0 if OK, negative if error (e.g., device not          */
/*     available, or a bus error)                                  */
/*                                                                 */
/* This function initializes the display, clears the screen buffer */
/* to white, writes the cleared buffer to the device, and sets     */
/* the configuration structure so the caller can determine the     */
/* device geometry and characteristics.                            */
/*                                                                 */
/* General state is initialized such that X=0, Y=0, the foreground */
/* colour is black and the background white.                       */
/* The DRAW_BOUNDS flag is cleared.                                */
/*                                                                 */
/* The text state is initialized so that the first character       */
/* written using drawPutc will appear at the top left of the       */
/* screen (by default, in the tollosFont5x7 font writing both      */
/* foreground and background).                                     */
/*                                                                 */
/* The screen buffer is cleared to white and the configuration and */
/* state are set even if an error occurred during device           */
/* initialization.                                                 */
/*                                                                 */
/* Note that it is not always possible to determine if the device  */
/* or screen is available, because some devices are 'write only'.  */
/*                                                                 */
/* The following items are reported in the display configuration   */
/* structure:                                                      */
/* * type      -- device type (binary, greyscale, or colour)       */
/* * width     -- screen width (pels)                              */
/* * height    -- screen height (pels)                             */
/* * depth     -- screen depth (bits/pel, maximum 32)              */
/* * xpitch    -- horizontal dot pitch (micrometers)               */
/* * ypitch    -- vertical dot pitch (micrometers)                 */
/* * writetime -- typical time to write a full screen of pels to   */
/* the physical device (microseconds); this is an estimate, not a  */
/* guarantee.                                                      */
/* --------------------------------------------------------------- */
int  drawInit(drawConfig *config) {
  int rc=drawInitIO(config);       // initialize device
  screen=*config;                  // save a copy

  // initialize general state
  state.drawX=0;                   // bottom left
  state.drawY=0;                   // ..
  drawPen(BLACK);                  // colours
  drawCanvas(WHITE);               // ..
  state.drawBounds=0;              // clear error flag

  // initialize text state
  state.textFont=&tollosFont5x7;   // current font
  state.textPen=1;                 // write all pels
  state.textCanvas=1;              // ..
  // the text position is set just off the bottom right of the screen
  // so that the first character write will cause a page wrap and set
  // the baseline correctly for the font active at that time
  state.textX=screen.width;        // current position
  state.textY=0;                   // ..

  return rc;
  } // drawInit

/* --------------------------------------------------------------- */
/* drawLine -- draw a line from current position                   */
/*                                                                 */
/*   x -- X coordinate to draw line to                             */
/*   y -- Y coordinate to draw line to                             */
/*                                                                 */
/* This draws a line from the current position to x, y, using the  */
/* pen colour; the current position is then updated to be x, y.    */
/* A compact version of Bresenham's algorithm is used.             */
/*                                                                 */
/* If x or y would not fit in 16 bits, DRAW_BOUNDS is set to 1 and */
/* no line is drawn.                                               */
/* --------------------------------------------------------------- */
void drawLine(int x, int y) {
  int cx, cy;                 // current x, y
  int xx, yy;                 // delta between from and to points
  int sx, sy;                 // step direction [1 or -1] in X & Y
  int dx, dy;                 // delta between points in X & Y
  int inc;                    // increment
  flag steep;                 // line is steep
  int i;                      // work
  if (x!=(int16)x || y!=(int16)y) {
    state.drawBounds=1;
    return;
    }
  cx=state.drawX;             // copy [may be swapped, below]
  cy=state.drawY;             // ..

  // draw the starting pel
  drawSetPelIO(cx, cy, penvalue);

  // Set up steps, deltas, and increment
  xx=x-cx;                    // change in X ..
  yy=y-cy;                    // .. and Y
  dx=abs(xx);                 // magnitudes thereof
  dy=abs(yy);                 // ..

  if (dx+dy==0) return;       // single pel -- no more to do

  sx=(xx>0) ? 1 : -1;         // step directions
  sy=(yy>0) ? 1 : -1;         // ..

  steep=(dy<=dx);             // note slope
  if (!steep) {               // not so steep; swap axes
    iswap(dx, dy);
    iswap(cx, cy);
    iswap(sx, sy);
    }
  inc=(dy*2)-dx;              // incremental change required

  // time to draw the line
  for (i=0; i<dx; i++) {      // to include final pel
    for(; inc>=0;) {
      cy+=sy;
      inc-=dx*2;
      }
    cx+=sx;
    inc+=dy*2;
    if (steep) drawSetPelIO(cx, cy, penvalue);
     else      drawSetPelIO(cy, cx, penvalue);
    } // i

  state.drawX=x;              // new state
  state.drawY=y;              // ..
  } // drawLine


/* --------------------------------------------------------------- */
/* drawOval -- draw an oval (ellipse) around the current position  */
/*                                                                 */
/*   a -- radius (semiaxis) in the X dimension (pels)              */
/*   b -- radius (semiaxis) in the Y dimension (pels)              */
/*                                                                 */
/* This draws an ellipse centred on the current position using the */
/* pen colour.  The ellipse will have its axes parallel to the X   */
/* and Y axes of the screen; its width will be 2*a+1 pels and its  */
/* height 2*b+1.  Hence, if a==b a circle (in pel space) is drawn. */
/*                                                                 */
/* To draw a circle when pels are not square, use the screen pitch */
/* values to scale the value of a or b.                            */
/*                                                                 */
/* If a or b is negative or greater than 896, DRAW_BOUNDS is set   */
/* and no oval is drawn.                                           */
/*                                                                 */
/* The algorithm used is that given in ''Getting Raster Ellipses   */
/* Right'', M. Douglas McIlroy, ACM Transactions on Graphics (TOG) */
/* Volume 11, No. 3, July 1992 (doi>10.1145/130881.130892).        */
/*                                                                 */
/* The arithmetic used in the algorithm fits in 32-bit integers    */
/* for values of a and b less than 896; confirmed experimentally.  */
/* --------------------------------------------------------------- */
#define incx() x++, dxt+=d2xt, t+=dxt
#define incy() y--, dyt+=d2yt, t+=dyt
void drawOval(int a, int b) {
  int x, y, a2, b2;                // as per McIlroy
  int crit1, crit2, crit3;         // ..
  int t, dxt, d2xt, dyt, d2yt;     // ..

  if ((uint)a>896 || (uint)b>896) {
    state.drawBounds=1;
    return;
    }

  x=0, y=b;
  a2=a*a, b2=b*b;
  crit1=-(a2/4+a%2+b2);
  crit2=-(b2/4+b%2+a2);
  crit3=-(b2/4+b%2);
  t=-a2*y;                         // t = e(x+1/2, y-1/2) - (a%2+b%2)/4
  dxt=2*b2*x, dyt=-2*a2*y;
  d2xt=2*b2, d2yt=2*a2;

  while(y>=0 && x<=a) {
    // debugf("x, y: %d, %d\n", x, y);
    drawSetPelIO(state.drawX+x, state.drawY+y, penvalue);
    if (x!=0 || y!=0) {
      drawSetPelIO(state.drawX-x, state.drawY-y, penvalue);
      }
    if (x!=0 && y!=0) {
      drawSetPelIO(state.drawX+x, state.drawY-y, penvalue);
      drawSetPelIO(state.drawX-x, state.drawY+y, penvalue);
      }

    if (t+b2*x <= crit1 ||         // e(x+1, y-1/2) <= 0
        t+a2*y <= crit3)           // e(x+1/2, y) <= 0
      incx();
     else if (t-a2*y > crit2)      // e(x+1/2, y-1) > 0
      incy();
     else {
      incx();
      incy();
      }
    } // while
  } // drawOval

/* --------------------------------------------------------------- */
/* drawPel -- set pel at current X and Y to pen or canvas colour   */
/*                                                                 */
/*   pel -- 0 to set pel to canvas, 1 to set to pen colour         */
/*                                                                 */
/* Only the screen buffer is affected.                             */
/* No error is possible.                                           */
/* --------------------------------------------------------------- */
void drawPel(flag pel) {
  drawSetPelIO(state.drawX, state.drawY, pel ? canvalue : penvalue);
  } // drawPel

/* --------------------------------------------------------------- */
/* drawPen -- set the pen (foreground) colour                      */
/*                                                                 */
/*   rgb -- colour value [0..255 in each]                          */
/*                                                                 */
/* This sets the Pen RGB value.                                    */
/* No error is possible.                                           */
/* --------------------------------------------------------------- */
void drawPen(drawRGB rgb) {
  // save value in lookaside and RGB in device context
  penvalue=drawFromRGBIO(rgb);     // convert to device value
  state.drawPen=rgb;
  } // drawPen

/* --------------------------------------------------------------- */
/* drawPutc -- display character on screen                         */
/*                                                                 */
/*   ch -- character to display                                    */
/*                                                                 */
/* The character is displayed using the current font, text X and   */
/* Y position, and pen and canvas colours.  The character          */
/* rectangle defined in the font is positioned such that its left  */
/* column will be at the text X position and its bottom left       */
/* corner will be d pels below the text Y position (where d is     */
/* the font descent -- the number of rows below the baseline).     */
/*                                                                 */
/* For each bit with a value of 1 in the character the             */
/* corresponding pel is written with the pen colour, provided that */
/* TEXT_PEN is 1.                                                  */
/* For each bit with a value of 0 in the character the             */
/* corresponding pel is written with the canvas colour, provided   */
/* that TEXT_CANVAS is 1.                                          */
/*                                                                 */
/* If either TEXT_PEN or TEXT_CANVAS is set, this function will    */
/* ensure a maximum width of the character is written to the       */
/* screen: if there is insufficient space to the right of the      */
/* character, the rest of the line is cleared (if TEXT_CANVAS is   */
/* set) and then the text X and Y positions are wrapped to the     */
/* next row (which may then wrap to the top of the screen).  The   */
/* character is then written.                                      */
/*                                                                 */
/* In all cases the X position is then incremented by the width of */
/* the character, +1 pel of spacing if FONT_SPACE (an attribute of */
/* the font) is set.                                               */
/*                                                                 */
/* If wrap to a new row takes place, leading of one pel is added   */
/* if FONT_LEAD is set.  Any spacing or leading will be written    */
/* with the canvas colour if TEXT_CANVAS is set.                   */
/*                                                                 */
/* If neither TEXT_PEN nor TEXT_CANVAS is set, no line wrap takes  */
/* place and so the change in X position will reflect the width    */
/* of the character (and any spacing) always; this can be used to  */
/* measure the width of characters, formatted strings, etc.        */
/*                                                                 */
/* An ASCII control character (code point < 0x20, or 0x7f) is only */
/* treated as a control character if the current font does not     */
/* define a character at that code point and either TEXT_PEN or    */
/* TEXT_CANVAS is set.  Control characters are ignored (have no    */
/* effect) except for the following:                               */
/* * CR: sets the X position to 0                                  */
/* * LF: move the Y position down a line (may cause a wrap to top  */
/* of the screen when the next character is written)               */
/* * FF: clear the screen and reset the X and Y positions so the   */
/* next character will appear at top left                          */
/* * TAB: moves the X position to the next tab position that is at */
/* least nwidth+FONT_SPACE pels to the right; a tab position is    */
/* textTab*(nwidth+FONT_SPACE) pels; this may cause a line wrap    */
/* when the next character is written.                             */
/*                                                                 */
/* If a character is not in a font and is not a control character  */
/* then the first character in the font will be displayed.         */
/*                                                                 */
/* Characters are permitted to touch all edges of the screen (it   */
/* is not required to have pel space to the right or below), even  */
/* if FONT_SPACE and/or FONT_LEAD are set.                         */
/*                                                                 */
/* The text X and Y positions are not necessarily on-screen;       */
/* however, they are limited to 32 bits.                           */
/* --------------------------------------------------------------- */
void drawPutc(int ch) {
  const byte *bp;                  // work
  uint mask;                       // ..
  int wid;                         // character width (pels)
  int x, y, yhi, ylo;              // pel coordinates
  int offset=ch-state.textFont->first; // character offset into data
  int spacing=(state.textFont->flags & FONT_SPACE)!=0;
  int leading=(state.textFont->flags & FONT_LEAD)!=0;

  // check character is in the font; if not, it may be a control
  // character and will be treated appropriately; if not a control
  // character then the first character in the font is substituted
  // (note here that chars is the count less 1, allowing 256 in a font)
  if (offset<0 || offset>state.textFont->chars) {
    if (ch<0x20 || ch==0x7f) {          // is control character?
      if (!state.textPen && !state.textCanvas); // nop unless inking
       else if (ch==ASCII_CR) state.textX=0;
       else if (ch==ASCII_LF) linefeed(leading);
       else if (ch==ASCII_FF) drawClearIO(canvalue);
       else if (ch==ASCII_TAB) {
        // tab rack positions are multiples of textTab en-spaces
        uint tabwidth=(state.textFont->nwidth+spacing)*textTab;
        uint x=state.textX+state.textFont->nwidth+spacing; // at least to here
        state.textX=((x+tabwidth-1)/tabwidth)*tabwidth;    // round down
        }
      // all other control characters are ignored
      return;                           // done
      } // controls
    offset=0;                           // is not control; map to first
    }

  // find the character
  bp=state.textFont->data+offset*state.textFont->bytes; // -> character data
  // determine the character width
  if (state.textFont->flags & FONT_MONOSPACE) wid=state.textFont->width;
   else {                               // FONT_PROPORTIONAL)
    wid=*bp;                            // width is first byte
    bp++;                               // step over it
    }

  if (state.textPen || state.textCanvas) { // if writing ink
    if (state.textX+wid > (int)screen.width) {    // line wrap needed
      if (state.textCanvas) {                     // if writing background ..
        // .. wipe to end of line (consider proportional font debris)
        ylo=state.textY-state.textFont->descent;  // current bottom row ..
        yhi=ylo+state.textFont->height-1;         // .. and top row
        for (y=yhi; y+leading>ylo; y--) {
          for (x=state.textX; x<(int)screen.width; x++) {
            drawSetPelIO(x, y, canvalue);
            }
          }
        }
      state.textX=0;
      linefeed(leading);                // [will change textY]
      }

    // ready to write character at textX and textY (the latter being
    // baseline); here, bp -> first byte of packed bits and wid is the
    // number of bits per row

    // consume bits, writing a pel when a bit is set in data
    // calculate the high and low bounds of the character
    mask=0x80;                          // bitmask
    ylo=state.textY-state.textFont->descent; // bottom row ..
    yhi=ylo+state.textFont->height-1;   // .. and top row

    for (y=yhi; y>=ylo; y--) {          // top to bottom
      for (x=state.textX; x<state.textX+wid; x++) {
        if (*bp & mask) {
          if (state.textPen) drawSetPelIO(x, y, penvalue);
          }
         else {
          if (state.textCanvas) drawSetPelIO(x, y, canvalue);
          }
        mask>>=1;                       // shift to next bit
        if (mask==0x00) {               // consumed all 8 bits
          bp++;                         // step to next data byte
          mask=0x80;                    // restart mask
          }
        } // x
      // if writing background write the following pel too [OK if off screen]
      if (state.textCanvas && spacing) drawSetPelIO(x, y, canvalue);
      } // y

    // if writing background, write the leading line below
    if (state.textCanvas && leading && ylo>0) {
      for (x=state.textX; x<=state.textX+wid; x++) drawSetPelIO(x, y, canvalue);
      }
    } // something to write

  state.textX+=wid+spacing;             // ready for next character
  } // drawPutc

/* --------------------------------------------------------------- */
/* drawQuery -- query a draw state item                            */
/*                                                                 */
/*   item -- a constant listed in tollosDraw.h                     */
/*                                                                 */
/*   returns current value of the item, or -1 if the item is       */
/*     unknown                                                     */
/*                                                                 */
/* Valid items are:                                                */
/*                                                                 */
/*   DRAW_X        -- drawing position                             */
/*   DRAW_Y        -- ..                                           */
/*   DRAW_PEN_R    -- pen colour                                   */
/*   DRAW_PEN_G    -- ..                                           */
/*   DRAW_PEN_B    -- ..                                           */
/*   DRAW_CANVAS_R -- canvas color                                 */
/*   DRAW_CANVAS_G -- ..                                           */
/*   DRAW_CANVAS_B -- ..                                           */
/*   DRAW_BOUNDS   -- bounds error bit                             */
/*                                                                 */
/*   TEXT_X        -- text write position                          */
/*   TEXT_Y        -- ..                                           */
/*   TEXT_PEN      -- 1 if 1 bits write (in pen colour)            */
/*   TEXT_CANVAS   -- 1 if 0 bits write (in canvas colour)         */
/*                                                                 */
/* Notes:                                                          */
/* # DRAW_X and DRAW_Y values will always fit in 16 bits           */
/* # Colour values(_R, _G, _B) have values in the range 0..255     */
/* # DRAW_BOUNDS, TEXT_PEN, and TEXT_CANVAS will be 0 or 1         */
/* # TEXT_X and TEXT_Y may have any 32-bit value                   */
/* --------------------------------------------------------------- */
int drawQuery(uint item)  {
  switch(item) {
    case DRAW_X:        return state.drawX;
    case DRAW_Y:        return state.drawY;
    case DRAW_PEN_R:    return state.drawPen.r;
    case DRAW_PEN_G:    return state.drawPen.g;
    case DRAW_PEN_B:    return state.drawPen.b;
    case DRAW_CANVAS_R: return state.drawCanvas.r;
    case DRAW_CANVAS_G: return state.drawCanvas.g;
    case DRAW_CANVAS_B: return state.drawCanvas.b;
    case DRAW_BOUNDS:   return state.drawBounds;

    case TEXT_X:        return state.textX;
    case TEXT_Y:        return state.textY;
    case TEXT_PEN:      return state.textPen;
    case TEXT_CANVAS:   return state.textCanvas;

    default: return -1;  // bad item
    } // switch
  } // drawQuery

/* --------------------------------------------------------------- */
/* drawSetFont -- set new current font                             */
/*                                                                 */
/*   font -- the font to be the new current font                   */
/*                                                                 */
/* The font is a filled-in and valid Tollos fontInfo structure,    */
/* with associated data.                                           */
/*                                                                 */
/* No other state (such as text X and Y position) is affected.     */
/* No error is possible.                                           */
/* --------------------------------------------------------------- */
void drawSetFont(const fontInfo *font) {
  state.textFont=font;
  } // drawSetFont

/* --------------------------------------------------------------- */
/* drawSetPel -- set a pel to a specific colour                    */
/*                                                                 */
/*   x   -- pel X coordinate                                       */
/*   y   -- pel Y coordinate                                       */
/*   rgb -- colour value [0..255 in each]                          */
/*                                                                 */
/* This sets the pel at x, y to the RGB value; if x or y would     */
/* not fit in 16 bits this has no effect other than to set         */
/* DRAW_BOUNDS to 1.                                               */
/* --------------------------------------------------------------- */
void drawSetPel(int x, int y, drawRGB rgb) {
  if (x!=(int16)x || y!=(int16)y) state.drawBounds=1;
   else drawSetPelIO(x, y, drawFromRGBIO(rgb));
  } // drawSetPel

/* --------------------------------------------------------------- */
/* drawStop -- stop the display device                             */
/*                                                                 */
/* This puts the device into a clean or low power/off mode, if     */
/* appropriate.                                                    */
/* For LCDs this should save power as well as properly discharging */
/* the display.                                                    */
/* No error is possible.                                           */
/* --------------------------------------------------------------- */
void drawStop(void) {
  drawStopIO();                    // simple wrapper
  } // drawStop

/* --------------------------------------------------------------- */
/* drawTextBits -- sets text drawing behavior                      */
/*                                                                 */
/*   canvas -- non-0 to write font 0 bits in canvas colour         */
/*   pen    -- non-0 to write font 1 bits in pen colour            */
/*                                                                 */
/* This sets the TEXT_CANVAS and TEXT_PEN flags which control      */
/* the writing of font 0 and 1 bits to the screen.  If both are    */
/* 0 no pels will be written but characters can be measured.       */
/*                                                                 */
/* See drawPutc for more detail.                                   */
/* No error is possible.                                           */
/* --------------------------------------------------------------- */
void drawTextBits(flag canvas, flag pen) {
  state.textCanvas=(canvas!=0);
  state.textPen=(pen!=0);
  } // drawTextBits

/* --------------------------------------------------------------- */
/* drawTextXY -- set Text X and Y position                         */
/*                                                                 */
/*   x -- X coordinate                                             */
/*   y -- Y coordinate                                             */
/*                                                                 */
/* This sets the text X and Y positions; x or y may be off-screen; */
/* this is not an error -- they will be brought on-screen when the */
/* next character is placed by drawPutc.                           */
/* --------------------------------------------------------------- */
void drawTextXY(int x, int y) {
  state.textX=x;
  state.textY=y;
  } // drawTextXY

/* --------------------------------------------------------------- */
/* drawWrite -- write buffer to screen (display to user)           */
/*                                                                 */
/*   returns 0 if OK, negative if error (e.g., bus error)          */
/*                                                                 */
/* This function sends the screen buffer to the device.  Depending */
/* on the device, this may send the whole buffer to the device,    */
/* or it may send only parts that have changed since the last call */
/* to drawWrite.                                                   */
/*                                                                 */
/* The DRAW_BOUNDS flag is cleared.                                */
/* --------------------------------------------------------------- */
int drawWrite(void) {
  state.drawBounds=0;              // clear error flag
  return drawWriteIO();            // effect the write
  } // drawWrite

/* --------------------------------------------------------------- */
/* drawXY -- set X and Y position                                  */
/*                                                                 */
/*   x -- X coordinate                                             */
/*   y -- Y coordinate                                             */
/*                                                                 */
/* This moves the current position to x, y; if x or y would not    */
/* fit in 16 bits, DRAW_BOUNDS is set to 1 and the current         */
/* position is not changed.                                        */
/* --------------------------------------------------------------- */
void drawXY(int x, int y) {
  if (x!=(int16)x || y!=(int16)y) state.drawBounds=1;
   else {state.drawX=x; state.drawY=y;}
  } // drawXY

/* --------------------------------------------------------------- */
/* linefeed -- step to next line, with possible screen wrap        */
/* --------------------------------------------------------------- */
static void linefeed(int leading) {
  // careful calculation for row wrap here as using uints
  if (state.textY < (int)state.textFont->height
                   +(int)state.textFont->descent+leading) {
    // back to top
    state.textY=screen.height-state.textFont->height+state.textFont->descent;
    }
   else state.textY-=state.textFont->height+leading; // down one char + leading
  } // linefeed
