/* --------------------------------------------------------------- */
/* tollosDraw.h -- draw (display) device functions                 */
/* --------------------------------------------------------------- */
/*                                                                 */
/* Copyright (c) Mike Cowlishaw, 2010-2011.  All rights reserved.  */
/*                                                                 */
/* Permission to use, copy, modify, and/or distribute this         */
/* software for any purpose with or without fee is hereby granted, */
/* provided that the above copyright notice and this permission    */
/* notice appear in all copies, and that notice and the date of    */
/* any modifications be added to the software.                     */
/*                                                                 */
/* This software is provided "as is".  No warranties, whether      */
/* express, implied, or statutory, including, but not limited to,  */
/* implied warranties of merchantability and fitness for a         */
/* particular purpose apply to this software.  The author shall    */
/* not, in any circumstances, be liable for special, incidental,   */
/* or consequential damages, for any reason whatsoever.            */
/*                                                                 */
/* --------------------------------------------------------------- */
/* The draw functions provide access to display device(s) (LCDs,   */
/* etc.), including text and graphics functions.                   */
/* --------------------------------------------------------------- */

#ifndef TOLLOS_DRAW_USED
#define TOLLOS_DRAW_USED

#include "tollos.h"           // types etc.
#include "tollosFont.h"       // font descriptors

/* Drawing and screen types */
// color structure
typedef struct {
  byte t;                     // attributes (not used by Tollos)
  byte r;                     // red
  byte g;                     // green
  byte b;                     // blue
  } drawRGB;

// display configuration structure (all values must be < 2**31)
typedef struct {
  uint type;                  // device type
  uint width;                 // screen width (pels)
  uint height;                // screen height (pels)
  uint depth;                 // screen depth (bits/pel, maximum 32)
  uint xpitch;                // horizontal dot pitch (micrometers)
  uint ypitch;                // vertical dot pitch (micrometers)
  uint writetime;             // typical time to write full screen (us)
  } drawConfig;

// drawConfig.type values
#define DRAW_TYPE_BINARY  1   // monochrome, one bit/pel
#define DRAW_TYPE_GREY    2   // monochrome (two or more bits/pel)
#define DRAW_TYPE_GRAY    2   // ..
#define DRAW_TYPE_COLOUR  3   // colourful
#define DRAW_TYPE_COLOR   3   // ..

// drawing and text state
typedef struct {
  // draw state
  int  drawX;                 // next write coordinate (may be offscreen)
  int  drawY;                 // ..
  drawRGB drawPen;            // Pen colour
  drawRGB drawCanvas;         // Pen colour

  // text state
  const fontInfo *textFont;   // current font
  int  textX;                 // next write coordinate (may be offscreen)
  int  textY;                 // ..

  // flags etc.
  flag textPen;               // 1 to draw pen pixels
  flag textCanvas;            // 1 to draw canvas pixels
  flag drawBounds;            // bounds error seen (an Y or Y too large)
  byte tbd;                   // [reserved]
  } drawState;

// public colours
extern const drawRGB BLACK;
extern const drawRGB CGA_DARKBLUE;
extern const drawRGB CGA_DARKGREEN;
extern const drawRGB CGA_DARKCYAN;
extern const drawRGB CGA_DARKRED;
extern const drawRGB CGA_DARKPINK;
extern const drawRGB CGA_BROWN;
extern const drawRGB PALEGREY;
extern const drawRGB DARKGREY;
extern const drawRGB CGA_BLUE;
extern const drawRGB CGA_GREEN;
extern const drawRGB CGA_CYAN;
extern const drawRGB CGA_RED;
extern const drawRGB CGA_PINK;
extern const drawRGB CGA_YELLOW;
extern const drawRGB WHITE;
// alternative spellings
#define PALEGRAY PALEGREY
#define DARKGRAY DARKGREY

/* Constants for drawQuery items */
#define DRAW_X           201  // drawing position
#define DRAW_Y           202  // ..
#define DRAW_PEN_R       203  // pen colour
#define DRAW_PEN_G       204  // ..
#define DRAW_PEN_B       205  // ..
#define DRAW_CANVAS_R    206  // canvas color
#define DRAW_CANVAS_G    207  // ..
#define DRAW_CANVAS_B    208  // ..
#define DRAW_BOUNDS      209  // bounds error flag

#define TEXT_X           221  // text write position
#define TEXT_Y           222  // ..
#define TEXT_PEN         223  // 1 if 1 bits write (in pen colour)
#define TEXT_CANVAS      224  // 1 if 0 bits write (in canvas colour)

/* --------------------------------------------------------------- */
/* Functions                                                       */
/* --------------------------------------------------------------- */
extern void drawBox(int x, int y, flag fill);
extern void drawCanvas(drawRGB rgb);
extern void drawClear(void);
extern int  drawf(const char *format, ...);
extern void drawFill(int x, int y);
extern void drawGetFont(const fontInfo **font);
extern void drawGetPel(int x, int y, drawRGB *rgb);
extern int  drawInit(drawConfig *config);
extern void drawLine(int x, int y);
extern void drawOval(int a, int b);
extern void drawPel(flag pel);
extern void drawPen(drawRGB rgb);
extern void drawPutc(int ch);
extern int  drawQuery(uint item);
extern void drawSetFont(const fontInfo *font);
extern void drawSetPel(int x, int y, drawRGB rgb);
extern void drawStop(void);
extern void drawTextBits(flag canvas, flag pen);
extern void drawTextXY(int x, int y);
extern int  drawWrite(void);
extern void drawXY(int x, int y);


#ifdef TOLLOS_IO
// Draw I/O functions [not for application use]

// actions
extern void drawClearIO(int pel);
extern int  drawGetPelIO(int x, int y);
extern int  drawInitIO(drawConfig *config);
extern void drawRawIO(byte *to, const byte *from);
extern void drawSetPelIO(int x, int y, int pel);
extern void drawStopIO(void);
extern int  drawWriteIO(void);

// conversions
extern int  drawFromRGBIO(drawRGB rgb);
extern void drawToRGBIO(int pel, drawRGB *rgb);


#endif

#endif
