/* --------------------------------------------------------------- */
/* tollosFont.h -- font header and flags                           */
/* --------------------------------------------------------------- */
/*                                                                 */
/* Copyright (c) Mike Cowlishaw, 2010-2011.  All rights reserved.  */
/*                                                                 */
/* Permission to use, copy, modify, and/or distribute this         */
/* software for any purpose with or without fee is hereby granted, */
/* provided that the above copyright notice and this permission    */
/* notice appear in all copies, and that notice and the date of    */
/* any modifications be added to the software.                     */
/*                                                                 */
/* This software is provided "as is".  No warranties, whether      */
/* express, implied, or statutory, including, but not limited to,  */
/* implied warranties of merchantability and fitness for a         */
/* particular purpose apply to this software.  The author shall    */
/* not, in any circumstances, be liable for special, incidental,   */
/* or consequential damages, for any reason whatsoever.            */
/*                                                                 */
/* --------------------------------------------------------------- */
/* The font definition here is used to describe bitmap fonts used  */
/* by the tollosShow text functions.                               */
/*                                                                 */
/* This is deliberately simplified, currently for left->right      */
/* text display, but has some potential for expansion; new items   */
/* can be added at the bottom of the fontInfo structure and flags  */
/* can be added.                                                   */
/*                                                                 */
/* The byte array pointed to by fontInfo.data has fontInfo.bytes   */
/* for each character in the array (the code for the first         */
/* character in the array is fontInfo.first).                      */
/*                                                                 */
/* If fontInfo.flags has FONT_PROPORTIONAL set then the first byte */
/* the character is its width in pels and the remaining bytes are  */
/* the bit data.                                                   */
/*                                                                 */
/* If fontInfo.flags has FONT_MONOSPACE set then the width of the  */
/* character is fontInfo.width and all the bytes of the character  */
/* are bit data.                                                   */
/*                                                                 */
/* Bit data are the bit image of the character, packed with each   */
/* row following the previous row, starting with the top row and   */
/* working down.                                                   */
/*                                                                 */
/* For example, the first two bytes would be 0x52 0xBE for the     */
/* monospace character:                                            */
/*                                                                 */
/*   01010                                                         */
/*   01010                                                         */
/*   11111                                                         */
/*   01010                                                         */
/*   11111                                                         */
/*   01010                                                         */
/*   01010                                                         */
/*                                                                 */
/* In proportional fonts some bytes may be unused -- however in    */
/* small fonts this is less wasteful than an index (and simpler).  */
/* --------------------------------------------------------------- */

#ifndef TOLLOS_FONT_USED
#define TOLLOS_FONT_USED

#include "tollos.h"           // types etc.

// font information structure; this must not be reordered
typedef struct {
  const char *name;           // -> name string
  const byte *data;           // -> first byte of bit data (packed)
  byte  flags;                // attributes (see below)
  byte  first;                // offset (code point) of first character
  byte  chars;                // count of characters - 1
  byte  bytes;                // bytes/character (including width)

  // metrics
  byte  width;                // character maximum width, pels
  byte  height;               // character total height, pels
  byte  nwidth;               // character n-width (en), pels
  byte  xheight;              // character x-height, pels
  byte  descent;              // character descent (pels below baseline)
  } fontInfo;

// font attributes, in fontInfo.flags
//
//     +--------+--------+
//     | x x x L x S P M |
//     +--------+--------+
//
// x = reserved, must be 0
// L = leading (1=add a row below character, 0=do not)
// S = spacing (1=add a pel after character, 0=do not)
// P = font is proportional
// M = font is monospaced
// (Only one of P and M shall be set)
//
#define FONT_MONOSPACE    0x01     // all characters fixed width
#define FONT_PROPORTIONAL 0x02     // first byte of each has width
#define FONT_SPACE        0x04     // add space
#define FONT_LEAD         0x10     // add lead

#endif

