/* --------------------------------------------------------------- */
/* tollosGyro.c -- gyroscope functions for Tollos                  */
/* --------------------------------------------------------------- */
/*                                                                 */
/* Copyright (c) Mike Cowlishaw, 2010-2011.  All rights reserved.  */
/*                                                                 */
/* Permission to use, copy, modify, and/or distribute this         */
/* software for any purpose with or without fee is hereby granted, */
/* provided that the above copyright notice and this permission    */
/* notice appear in all copies, and that notice and the date of    */
/* any modifications be added to the software.                     */
/*                                                                 */
/* This software is provided "as is".  No warranties, whether      */
/* express, implied, or statutory, including, but not limited to,  */
/* implied warranties of merchantability and fitness for a         */
/* particular purpose apply to this software.  The author shall    */
/* not, in any circumstances, be liable for special, incidental,   */
/* or consequential damages, for any reason whatsoever.            */
/*                                                                 */
/* --------------------------------------------------------------- */
/* These functions provide basic gyroscope functions for           */
/* Tollos, intended to be device-independent (although at the      */
/* moment an I2C connection is assumed).                           */
/*                                                                 */
/* To use, the device must be initialized using gyroInit.  This    */
/* indicates which I2C channel the device is on and which port/pin */
/* is used for the DataReady interrupt.  The gyroscope range is    */
/* determined by the device used.  (For example, for the ITG3200   */
/* it is +/- 2000 degrees/sec.)                                    */
/*                                                                 */
/* The latest gyroscope reading can be read safely using           */
/* gyroGetXYZ, which returns the three axes in milli-degrees/sec.  */
/* The axis directions are as defined by the device manufacturer.  */
/* Note that gyroscopes will generally have some constant offset,  */
/* or drift, that will indicate a rate of turn on a 'static'       */
/* device; this may be partly compensated for by adjust for the    */
/* die temperature, available using gyroTemperature.               */
/*                                                                 */
/* The default update rate is selected by the device driver code   */
/* and will be at least 25Hz (that is, the data will be no more    */
/* than 40ms old).  Update is not necessarily synchronous with the */
/* Tollos tick rate.   The rate can be changed with gyroRate; the  */
/* low pass-filter bandwidth is set to match or exceed the rate.   */
/*                                                                 */
/* --------------------------------------------------------------- */

#define  TOLLOS_IO            // allow use of I/O functions
#include "tollos.h"

// vector of readings from Gyroscope
typedef struct {
  int x;                      // X reading [mg]
  int y;                      // Y reading [mg]
  int z;                      // Z reading [mg]
  } gyroData;

// Double-buffer of readings; this is used to avoid race conditions
// between the interrupt handler and the gyroGetXYZ function as
// follows:
//
//   1. On entry to the Get function, gyroLast is flipped 0<->1.
//      (It is assumed that the interrupt handler cannot be running
//      at the moment gyroLast is actually written, but even if it
//      were, the buffer chosen would be arbitrary but would be
//      wholly written before the Get function regained control.)
//
//   2. The Get function tests gyroGoodData[gyroLast]; if 0, no good
//      data is available and this is reported to the caller.
//
//   3. If a gyroBuff does have good data this is set to the caller;
//      then the gyroGoodData flag is cleared.
//
//   4. In the interrupt handler, data are read from the device and
//      written to the non-gyroLast buffer, then gyroGoodData for that
//      buffer is set to 1 (this may happen many times, overwriting
//      the data, until there is a read).
//
// Hence: the two levels of code always write to and read from
// different buffers; access to those buffers is controlled by
// gyroLast alone, which is only switched in the lower priority
// task (when it is not accessing the buffer).
//
static gyroData gyroBuff[2];       // double buffer
static flag     gyroGoodData[2];   // 1 if the corresponding gyroBufff
                                   // has good data
static uint     gyroLast=0;        // which buffer was last read


// local function
static void gyroHandler(void);     // interrupt handler

/* --------------------------------------------------------------- */
/* gyroGetXYZ -- read latest x, y, and z angular rates             */
/*                                                                 */
/*   x, y, z -- receive angular rates [milli-degrees/sec:          */
/*              1000 = 1.000 degrees/second]; set to 0 if error    */
/*                                                                 */
/*   returns 0 if data good, -1 if error (no good data since last  */
/*     read, or device not initialized)                            */
/*                                                                 */
/* Results are returned scaled to milli-degrees/second.            */
/*                                                                 */
/* This must not be called from an interrupt level with higher     */
/* priority than the gyroscope interrupt.                          */
/* --------------------------------------------------------------- */
int gyroGetXYZ(int *x, int *y, int*z) {
  // see above for the handshaking protocol used here
  gyroLast=!gyroLast;
  if (!gyroGoodData[gyroLast]) {   // no good data
    *x=0; *y=0; *z=0;
    return -1;
    }

  // return data from the 'good' buffer
  *x=gyroBuff[gyroLast].x;
  *y=gyroBuff[gyroLast].y;
  *z=gyroBuff[gyroLast].z;

  gyroGoodData[gyroLast]=0;        // indicate data used

  return 0;                        // is good
  } // gyroGetXYZ

/* --------------------------------------------------------------- */
/* gyroHandler -- interrupt handler, called when Data Ready        */
/*                                                                 */
/* This reads the data into the 'active' buffer, then when and if  */
/* good sets gyroGoodData=1 for that buffer.                       */
/* --------------------------------------------------------------- */
static void gyroHandler(void) {
  int rc;                          // work
  int read=!gyroLast;              // buffer to read into

  rc=gyroGetXYZIO(&gyroBuff[read].x, &gyroBuff[read].y, &gyroBuff[read].z);
  // interrupt is cleared by the read
  if (rc<0) return;                // bad data

  gyroGoodData[read]=1;            // have goodness
  } // gyroHandler

/* --------------------------------------------------------------- */
/* gyroInit -- configure and initialize gyroscope                  */
/*                                                                 */
/*   channel -- I2C channel on which device is to be found         */
/*   drport  -- port for the Data Ready interrupt pin              */
/*   drpin   -- pin number (not bit mask) for the Data Ready       */
/*              interrupt pin                                      */
/*                                                                 */
/*   returns 0 if initialized OK, negative if error                */
/*                                                                 */
/* This sets up the pin and interrupt handler for the Data Ready   */
/* interrupt from the device then calls the device driver          */
/* initialization function to start the device.                    */
/*                                                                 */
/* The gyroInitIO function will set:                               */
/* * an update rate of at least 25 Hz (but not necessarily higher  */
/* than that, to save power); this may be changed with gyroRate    */
/* * an interrupt on Data Ready.                                   */
/*                                                                 */
/* Note that it can be 50ms or more before the first valid data    */
/* are available, so a gyroGetXYX call before that time may report */
/* no data available, as might a call that reads the device less   */
/* than 25ms after the last read.                                  */
/* --------------------------------------------------------------- */
int gyroInit(uint channel, uint drport, uint drpin) {
  int rc;                     // work

  gyroLast=0;                 // must be 0 or 1
  gyroGoodData[0]=0;          // no good data yet
  gyroGoodData[1]=0;          // ..

  // set up interrupt pin from number (bitFuns use a mask)
  rc=bitEnable(drport, 1<<drpin, BIT_RISE, gyroHandler);

  if (rc==0) rc=gyroInitIO(channel);
  return rc;
  } // gyroInit

/* --------------------------------------------------------------- */
/* gyroRate -- change update rate of gyroscope                     */
/*                                                                 */
/*   rate -- rate requested [Hz, 1..200]                           */
/*                                                                 */
/*   returns 0 if rate changed OK, -1 if error (rate zero or       */
/*     too high)                                                   */
/*                                                                 */
/* The gyroscope will be set to a slowest available rate that      */
/* is >= the requested rate, to minimize resource usage.  The      */
/* low-pass filter bandwidth is set to match or exceed the rate.   */
/*                                                                 */
/* To ensure working with 100kHz I2C, the maximum rate is limited  */
/* to 200Hz -- but note that this may use significant processor    */
/* bandwidth.                                                      */
/* --------------------------------------------------------------- */
int gyroRate(uint rate) {
  return gyroRateIO(rate);    // pass on and back
  } // gyroRate

/* --------------------------------------------------------------- */
/* gyroStop -- stop and power-down gyroscope                       */
/*                                                                 */
/*   returns 0 if gyroscope stopped OK, negative if an error       */
/*     or it was not running.                                      */
/* --------------------------------------------------------------- */
int gyroStop(void) {
  int rc=gyroStopIO();        // stop the device
  gyroGoodData[0]=0;          // invalidate any data
  gyroGoodData[1]=0;          // ..
  return rc;
  } // gyroStop

/* --------------------------------------------------------------- */
/* gyroTemperature -- return current gyroscope die temperature     */
/*                                                                 */
/*   returns the latest temperature reading in milli-degrees       */
/*     Celsius, or GYRO_NOTEMP if no reading is available          */
/*                                                                 */
/* Many gyroscope devices provide the temperature of the chip die, */
/* which can be used to compensate for drift due to temperature    */
/* change.  GYRO_NOTEMP is returned if no temperature reading is   */
/* available or the Gyro device is stopped.                        */
/* --------------------------------------------------------------- */
int gyroTemperature(void) {
  return gyroTemperatureIO(); // pass back
  } // gyroTemperature

