/* --------------------------------------------------------------- */
/* tollosI2C.c -- I2C bus functions for Tollos                     */
/* --------------------------------------------------------------- */
/*                                                                 */
/* Copyright (c) Mike Cowlishaw, 2010-2011.  All rights reserved.  */
/*                                                                 */
/* Permission to use, copy, modify, and/or distribute this         */
/* software for any purpose with or without fee is hereby granted, */
/* provided that the above copyright notice and this permission    */
/* notice appear in all copies, and that notice and the date of    */
/* any modifications be added to the software.                     */
/*                                                                 */
/* This software is provided "as is".  No warranties, whether      */
/* express, implied, or statutory, including, but not limited to,  */
/* implied warranties of merchantability and fitness for a         */
/* particular purpose apply to this software.  The author shall    */
/* not, in any circumstances, be liable for special, incidental,   */
/* or consequential damages, for any reason whatsoever.            */
/*                                                                 */
/* --------------------------------------------------------------- */
/* Functions to read and write data from I2C bus devices, in I2C   */
/* master mode (the microcontroller is the Master):                */
/* * Get functions to read one or more bytes from registers        */
/* (using the convention that a data byte is first written to      */
/* identify the register)                                          */
/* * Put functions to write one or more bytes to registers         */
/* (using the convention that a data byte is first written to      */
/* identify the register)                                          */
/* * a PutGet function to write an arbitrary number of data bytes  */
/* followed by a read of an arbitrary number of bytes (this is     */
/* the 'low level' I/O supported by I2C)                           */
/* * a test/debug function to scan an I2C channel for devices.     */
/*                                                                 */
/* Where these functions take a 'device' parameter, this is the    */
/* i2c channel number (bus interface, 0, 1, 2, etc.) combined with */
/* the 7-bit I2C physical address by shifting the channel number   */
/* left 16 bits and ORing it with the 7-bit device physical        */
/* address.  The 16-bit shift allows for 10-bit I2C device         */
/* addresses; however the latter are not yet supported.            */
/* --------------------------------------------------------------- */

#define  TOLLOS_IO            // allow use of I/O functions
#include "tollos.h"

/* --------------------------------------------------------------- */
/* i2cGetReg -- read a byte from a register                        */
/*                                                                 */
/*   device -- channel and 7-bit physical address of the device    */
/*   reg    -- register (start address) to be read                 */
/*   get    -- place to store the byte read                        */
/*                                                                 */
/*   returns 0 if OK, negative if an error occurred                */
/* --------------------------------------------------------------- */
int i2cGetReg(uint device, byte reg, byte *get) {

  // write one byte address then read 1 byte data
  return i2cIO(device, &reg, 1, get, 1);
  } // i2cGetReg

/* --------------------------------------------------------------- */
/* i2cGetRegs -- read bytes from registers                         */
/*                                                                 */
/*   device -- channel and 7-bit physical address of the device    */
/*   reg    -- register (start address) to be read                 */
/*   get    -- place to store the bytes read                       */
/*   len    -- number of bytes to be read                          */
/*                                                                 */
/*   returns 0 if OK, negative if an error occurred                */
/* --------------------------------------------------------------- */
int i2cGetRegs(uint device, byte reg, byte *get, uint len) {

  // write one byte address then read len bytes data
  return i2cIO(device, &reg, 1, get, len);
  } // i2cGetRegs

/* --------------------------------------------------------------- */
/* i2cPutGet -- general I2C low-level O/I function                 */
/*                                                                 */
/*   device -- channel and physical address of device              */
/*             (channel shifted left 16 bits and ORed with 7-bit   */
/*             or 10-bit I2C physical address); currently only     */
/*             7-bit addresses are supported                       */
/*   put    -- start of byte(s) to write                           */
/*   putlen -- count of byte(s) to write [may be 0]                */
/*   get    -- start of byte(s) to read                            */
/*   getlen -- count of byte(s) to read [may be 0]                 */
/*                                                                 */
/*   returns 0 if OK, negative if an error occurred                */
/*                                                                 */
/* put is ignored if putlen is 0 (NULL recommended).               */
/* get is ignored if getlen is 0 (NULL recommended).               */
/*                                                                 */
/* This function effects a lower-level I2C bus operation.  First,  */
/* putlen bytes are written to the device.  Next, the device is    */
/* addressed with a request to read, and getlen bytes are read.    */
/*                                                                 */
/* For many devices the first data byte written to the device is   */
/* a register address, and then a continuing write writes to that  */
/* register and a read reads from that register.                   */
/* --------------------------------------------------------------- */

int i2cPutGet(uint device, byte *put, uint putlen, byte *get, uint getlen) {
  return i2cIO(device, put, putlen, get, getlen);
  } // i2cPutGet
/* --------------------------------------------------------------- */
/* i2cPutVal -- write a byte                          */
/*                                                                 */
/*   device -- channel and 7-bit physical address of the device    */
/*               */
/*   put    -- byte to be written                                  */
/*                                                                 */
/*   returns 0 if OK, negative if an error occurred                */
/* --------------------------------------------------------------- */
int i2cPutVal(uint device,byte Val) {
  byte buff[1];          // work buffer

  buff[0]=Val;           // Value ..
  
  return i2cIO(device, buff, 1, NULL, 0);
  } // i2cPutVal

/* --------------------------------------------------------------- */
/* i2cPutReg -- write a byte to a register                         */
/*                                                                 */
/*   device -- channel and 7-bit physical address of the device    */
/*   reg    -- register (start address) to be written              */
/*   put    -- byte to be written                                  */
/*                                                                 */
/*   returns 0 if OK, negative if an error occurred                */
/* --------------------------------------------------------------- */
int i2cPutReg(uint device, byte reg, byte put) {
  byte buff[2];          // work buffer

  buff[0]=reg;           // addr ..
  buff[1]=put;           // .. followed by data

  return i2cIO(device, buff, 2, NULL, 0);
  } // i2cPutReg

/* --------------------------------------------------------------- */
/* i2cPutRegs -- write bytes to registers                          */
/*                                                                 */
/*   device -- channel and 7-bit physical address of the device    */
/*   reg    -- register (start address) to be written              */
/*   put    -- address of bytes to be written                      */
/*   len    -- number of bytes to be written                       */
/*                                                                 */
/*   returns 0 if OK, negative if an error occurred                */
/* --------------------------------------------------------------- */
int i2cPutRegs(uint device, byte reg, byte *put, uint len) {
  byte buff[len+1];      // work buffer
  uint u;                // work

  buff[0]=reg;                               // addr ..
  for (u=0; u<len; u++) buff[u+1]=put[u];    // .. followed by data

  return i2cIO(device, buff, len+1, NULL, 0);
  } // i2cPutRegs


/* --------------------------------------------------------------- */
/* i2cScan -- scan I2C channel for devices                         */
/*                                                                 */
/*   device1 -- device address of first device to scan             */
/*   device2 -- device address of last device to scan [>= device1] */
/*                                                                 */
/*   returns count of devices found, or -1 if an error             */
/*                                                                 */
/* Both device addresses must be in the same channel (i.e., their  */
/* most significant 16 bits are identical).                        */
/*                                                                 */
/* This scan reads 'register 0x00' at each device address in the   */
/* requested range, and reports (using debugf) any device that     */
/* responds to that read.                                          */
/*                                                                 */
/* By convention, many devices return a 'who am I' value for a     */
/* read from register 0x00, however -- because a read of a register*/
/* is itself a convention (write register then read a byte) -- the */
/* results of this scan should be interpreted carefully.           */
/* --------------------------------------------------------------- */
int i2cScan(uint device1, uint device2) {
  uint dev;                             // work
  byte whoami;                          // ..
  int  num=0;                           // devices found
  uint channel=i2cChannel(device1);     // channel in use

  if (channel!=i2cChannel(device2)) return -1;
  if (device1>device2) return -1;

  // debugf("Scanning i2c%d ...\n", channel);
  for (dev=device1; dev<=device2; dev++) {
    // read register 0
    if (i2cGetReg(dev, 0x00, &whoami)==0) {
      debugf("i2c%d address %02x reports %02x\n",
             channel, i2cAddr7Bit(dev), whoami);
      num++;
      }
    } // dev
  //debugf("Scanned i2c%d ...\n", num);

  return num;
  } // i2cScan

