/* --------------------------------------------------------------- */
/* tollosUtil.c -- useful macros and functions for Tollos          */
/* --------------------------------------------------------------- */
/*                                                                 */
/* Copyright (c) Mike Cowlishaw, 2010-2011.  All rights reserved.  */
/*                                                                 */
/* Permission to use, copy, modify, and/or distribute this         */
/* software for any purpose with or without fee is hereby granted, */
/* provided that the above copyright notice and this permission    */
/* notice appear in all copies, and that notice and the date of    */
/* any modifications be added to the software.                     */
/*                                                                 */
/* This software is provided "as is".  No warranties, whether      */
/* express, implied, or statutory, including, but not limited to,  */
/* implied warranties of merchantability and fitness for a         */
/* particular purpose apply to this software.  The author shall    */
/* not, in any circumstances, be liable for special, incidental,   */
/* or consequential damages, for any reason whatsoever.            */
/*                                                                 */
/* --------------------------------------------------------------- */
/*                                                                 */
/* The macros are defined in ""tollosUtil.h"", and are:            */
/*                                                                 */
/*   abs(x)     -- return the absolute value of x                  */
/*   iswap(x,y) -- swap the values of the ints x and y using XOR   */
/*   max(a,b)   -- return the larger of x and y                    */
/*   min(a,b)   -- return the smaller of x and y                   */
/*   signum(x)  -- return the sign of x (-1, 0, or 1)              */
/*                                                                 */
/* The functions have various purposes.                            */
/* --------------------------------------------------------------- */

#include "tollos.h"

// if minidec is not available then define powers of 10 array here
#ifndef MINIDEC
  // Powers of ten (powers[n]==10**n, 0<=n<=9)
  const int powers[10]={1, 10, 100, 1000, 10000, 100000, 1000000,
                        10000000, 100000000, 1000000000};
#endif

/* --------------------------------------------------------------- */
/* byte2hex -- convert a byte to two hex chars                     */
/*                                                                 */
/*   b       -- byte to convert                                    */
/*   hex     -- points to where ASCII bytes should go              */
/*   upper   -- 1 to use ABCDEF, 0 for abcdef                      */
/*                                                                 */
/* This writes only the two characters; no terminating zero char   */
/* is added.  No error is possible.                                */
/* --------------------------------------------------------------- */
void byte2hex(byte b, char *hex, flag upper) {
  static const char *uppers="0123456789ABCDEF";
  static const char *lowers="0123456789abcdef";
  const char *letters=(upper ? uppers : lowers);  // select table

  hex[0]=letters[(unsigned)b>>4];
  hex[1]=letters[b & 0x0f];
  } // byte2hex

/* --------------------------------------------------------------- */
/* fastrand -- return a random number in the range 0 to n-1        */
/*                                                                 */
/*   n    -- defines the random number range                       */
/*   seed -- static or persistent seed                             */
/*                                                                 */
/*   returns the pseudo-random number                              */
/*                                                                 */
/* n should be <30,000,000; the seed is provided by the caller     */
/* initially but should not then be altered by the caller.         */
/*                                                                 */
/* This uses the conventional  X(n+1)=mod((A*X(n) + C),M) as used  */
/* in the Rexx function package, but without some refinements.     */
/*                                                                 */
/* The values for A and M are from the table in D.E.Knuth's        */
/* book, "Seminumerical Algorithms", Addison-Wesley, 1981,         */
/* page 102 (section 3.3.2).                                       */
/* These particular values are attributed to Lavaux and Janssens.  */
/* * C is chosen as 1                                              */
/* * M is 2**32                                                    */
/* * A is 1664525      (RNDFACT)                                   */
/*                                                                 */
/* --------------------------------------------------------------- */
int fastrand(uint n, uint *seed) {
  *seed=(*seed)*1664525L+1;        // don't care about overflow
  return (int)((*seed>>3)%n);      // return 0->N-1
  } // fastrand


/* --------------------------------------------------------------- */
/* hex2byte -- convert two hexadecimal characters to a byte        */
/*                                                                 */
/*   hex     -- points to two ASCII bytes                          */
/*   b       -- points to byte which will receive the converted    */
/*              hexadecimal; set to 0x00 if error                  */
/*                                                                 */
/*   returns 0 if OK, -1 if error                                  */
/*                                                                 */
/* This accepts both uppercase and lowercase A->F and a->f.        */
/* --------------------------------------------------------------- */
int hex2byte(char *hex, byte *b) {
  char *end;        // end char +1
  *b=0;
  for (end=hex+2; hex<end; hex++) {
    *b<<=4;
    if (*hex>='0' && *hex<='9') *b|=(*hex-'0');
    else if (*hex>='A' && *hex<='F') *b|=(*hex-'A'+10);
    else if (*hex>='a' && *hex<='f') *b|=(*hex-'a'+10);
    else {*b=0x00; return -1;}
    }
  return 0;
  } // hex2byte

/* --------------------------------------------------------------- */
/* isqrtg -- return the square root of an int, with guess          */
/*                                                                 */
/*   num   -- the number to square root                            */
/*   guess -- is the number to converge from (e.g., last similar   */
/*            result); if < 1, isqrtg will make the first guess    */
/*                                                                 */
/*   returns the integer square root of num, such that             */
/*     result*result < num; if num is less than zero then 0 is     */
/*     returned                                                    */
/*                                                                 */
/* Note: the maximum input value is 2**31-1 (2147483647)           */
/* hence the maximum possible returned value is 46340.             */
/* --------------------------------------------------------------- */
// mfc, derived from PalmGlobe lsqrtg
int isqrtg(int num, int guess) {
  int new, old;

  if (num<9) {                /* special case low values for safety */
    if (num<=0) return 0;     /* special case + safety */
    if (num< 4) return 1;     /* 1, 2, 3 */
    return 2;}                /* 4-8 */

  /* Note: first guess has dramatic performance impact -- a value   */
  /* 1, or N/16, both doubled total GLOBE draw time.                */
  /* N/1024+1, and 32,000 both worked well.  Dynamic better.        */
  /* dynamically calculated first guess, which is (just) better.    */
  old=guess;                  /* The first guess is a given */
  if (old<=0) old=32001;
  for (;;) {                  /* forever */
    new=(signed int)((unsigned int)(num/old+old)>>1); /* new guess */
    if ((new-old)<2
     && (new-old)>-2) break;  /* Done! */
    old=new;
    }
  return new;
  } // isqrtg

/* --------------------------------------------------------------- */
/* i2str -- layout an integer as a string                          */
/*                                                                 */
/*   string -- where to layout the integer                         */
/*   num    -- the integer to lay out (not 0x80000000)             */
/*   len    -- width of the field (string) to lay out into         */
/*                                                                 */
/*   returns 0 if OK, 1 if overflow (does not fit)                 */
/*                                                                 */
/* This writes num right-aligned and signed in the string field.   */
/* No terminating zero char is written.                            */
/* --------------------------------------------------------------- */
int i2str(char *string, int num, int len) {
  char *c=string+len-1;                 // -> rightmost char
  char sign='+';
  int i;
  if (num<0) {num=-num; sign='-';}      // broken if 0x80000000
  for (i=0; i<len; i++) string[i]=' ';  // clear
  for (;;) {
    if (c==string) {                    // no room for sign
      for (i=0; i<len; i++) string[i]='*'; // show
      return 1;
      }
    *c=(num%10)+'0';
    num=num/10;
    c--;
    if (num==0) break;
    }
  *c=sign;
  return 0;
  } // i2str

/* --------------------------------------------------------------- */
/* nop -- do nothing with an int                                   */
/*                                                                 */
/*   i -- an integer                                               */
/*                                                                 */
/*   returns i                                                     */
/*                                                                 */
/* This is used for placating compilers when parameters are        */
/* intentionally or temporarily unused.                            */
/* --------------------------------------------------------------- */
int nop(int i) {
  return i;
  } // nop

/* --------------------------------------------------------------- */
/* streq -- caseless string compare                                */
/*                                                                 */
/*   char1 -- first string to compare                              */
/*   char2 -- second string                                        */
/*                                                                 */
/*   returns 1 if the strings caseless-compare equal, 0 otherwise  */
/*                                                                 */
/* If both strings are empty they are equal.                       */
/* --------------------------------------------------------------- */
int streq(const char *char1, const char *char2) {
  for (;;char1++, char2++) {
    if (*char1==*char2) {if (*char1=='\0') return 1;}
     else if (toupper(*char1)!=toupper(*char2)) return 0;
    } // forever
  } // streq

/* --------------------------------------------------------------- */
/* strf -- a reentrant mini printf to send a formatted string      */
/*                                                                 */
/*   putC   -- the function to which to send formatted characters  */
/*   format -- format and data string                              */
/*   args   -- list of other arguments as required                 */
/*                                                                 */
/*   returns the number (count) of characters sent                 */
/*                                                                 */
/* The following format indicators are supported:                  */
/*                                                                 */
/*   %%    -- '%'                                                  */
/*   %c    -- single char                                          */
/*   %d, i -- signed int                                           */
/*   %m    -- minidec decimal number  [iff MINIDEC is defined]     */
/*   %s    -- string                                               */
/*   %u    -- unsigned int                                         */
/*   %x, X -- hexadecimal                                          */
/*   %\n   -- LF only (otherwise ""\n"" becomes CRLF)              */
/*                                                                 */
/* The five ""printf"" format flags (""+ - 0 #"" and space) are    */
/* supported for all formats as in C99 (e.g., ""#"" only affects   */
/* 'x', here).                                                     */
/*                                                                 */
/* The 'h', 'l', and 'L' modifiers are not supported.              */
/*                                                                 */
/* Field width (e.g., %10i, %08x) is supported for all formats.    */
/*                                                                 */
/* Field precision (e.g., %5.7i) is not supported; for decimals,   */
/* use Quantize on a number to set its 'precision'.                */
/*                                                                 */
/* A ""\n"" in the string is expanded to ""\r\n"" unless preceded  */
/* by ""%"".                                                       */
/*                                                                 */
/* Almost no error checking is carried out, so, for example,       */
/* the flags and field width can be used even if the format        */
/* character is not one of those listed; the format character then */
/* appears in the output.                                          */
/* --------------------------------------------------------------- */
// This code is reentrant
int strf(putter putC, const char *format, va_list args) {
  const char *f;         // pointer in format string
  int  count=0;          // chars put counter
  int  width;            // field width [0 means none, may be -ve]
  char sign;             // '+' flag, ' ' if space flag, 0x00 otherwise
  flag left;             // 1 if '-' left-justify flag
  flag alt;              // 1 if '#' flag
  char pad;              // '0' if '0' flag
  char buff[25];         // work, >17 for decimal, >24 for double 'a'
  char *source;          // where to get data from (buff, except for %s)
  int  sourlen;          // characters present at source

  for (f=format; *f!=0; f++) {
    if (*f!='%') {                           // not escape
      if (*f=='\n') {putC('\r'); count++;}   // LF -> CRLF
      putC(*f); count++;
      continue;
      }
    // have %
    f++;

    left=0; sign=0x00; alt=0; pad=' ';
    for (;; f++) {           // handle leading flags in any order
      if (*f=='-') left=1;         // left align in field
       else if (*f=='+') sign='+'; // use '+' on signed numbers
       else if (*f=='0') pad='0';  // pad numerics with 0
       else if (*f=='#') alt=1;    // alternate format for x and X
       else if (*f==' ') {         // use ' ' for '+'
        if (!sign) sign=' ';       // does not override +
        }
       else break;
      } // flags loop
    if (left) pad=' ';             // ignore '0' if '-' (C99 spec.)

    width=0;                       // gather width, if any
    for (; *f>='0' && *f<='9'; f++) width=width*10+(*f-'0');
    // [dot and precision could be gathered here]

    // now the data-dependent formatting; the numeric formats and %c
    // lay out the result in buff and set the length in sourlen; this
    // is then output (after padding) as necessary at the end
    //
    // %s changes the source pointer and also sets sourlen
    source=buff;                   // default source

    switch(*f) {                   // format indicator

      case 'c': {
        char c=(char)va_arg(args, int);
        buff[0]=c; sourlen=1;
        pad=' ';                   // not a signed format
        break;}

      case 'u':
      case 'd':
      case 'i': {
        uint u=(uint)va_arg(args, uint);
        char *bp=buff;   // work
        int  i;          // work
        flag had;        // had leading digit

        if (*f=='u') {   // unsigned
          if (sign) {*bp=sign; bp++;}
          }
         else {          // signed
          int i=(int)u;
          if (i>=0) {
            if (sign) {*bp=sign; bp++;}
            }
           else {
            *bp='-'; bp++;
            if (u==0x80000000) u=2147483648U;
             else u=-i;
            }
          }
        // u is now abs(arg), such that 0 <= u <= 4294967295 (2**32-1)
        had=0;
        if (u==0)  {*bp='0'; bp++;}
         else for (i=9; i>=0; i--) {
          uint d=u/powers[i];
          if (d) had=1;  // had leading digit
          u%=powers[i];
          if (had) {*bp='0'+d; bp++;}
          }
        sourlen=bp-buff;
        break;}

      #ifdef MINIDEC
      case 'm': {
        dec num=va_arg(args, dec);
        char *bp=buff;   // work

        if (sign && !decIsSigned(num)) {*bp=sign; bp++;}
        decToString(num, bp);
        sourlen=0;       // set length ..
        for (bp=buff; *bp; bp++) sourlen++;  // strlen
        break;}
      #endif

      case 's': {
        char *cp=(char *)va_arg(args, int);
        sourlen=0;
        if (cp!=NULL) {
          source=cp;
          for (; *cp; cp++) sourlen++;       // strlen
          }
        // [else use zero-length 'string' from buff]
        pad=' ';                   // not a signed format
        break;}

      case 'x':
      case 'X': {
        uint u=(uint)va_arg(args, uint);
        flag upper=(*f=='X');
        flag had;
        char x[8];
        char *bp=buff;   // work
        int i;

        // handle any 0x or 0X prefix: if padding with '0' and right-
        // aligning then it must be sent now; otherwise add to buffer
        if (alt) {
          if (pad=='0' && !left) {
            putC('0'); putC(*f);
            count+=2;
            width-=2;    // (may push it negative)
            }
           else {
            *bp='0'; bp++;
            *bp=*f;  bp++;
            }
          } // alt

        // convert whole to hex
        for (i=6; i>=0; i-=2) {
          byte2hex(u & 0xff, &x[i], upper);
          u>>=8;
          }
        // copy to buff without leading zeros
        had=0;
        for (i=0; i<8; i++) {
          if (!had && x[i]=='0' && i!=7) continue; // skip leading 0s
          had=1;
          *bp=x[i]; bp++;
          }
        sourlen=bp-buff;
        // [this can be padded with '0' but never has sign]
        break;}

      default: {                   // %? => %?  (so %\n => LF, %% => %)
        // all of these may have padding
        buff[0]=*f; sourlen=1;
        break;}
      } // format char

    // all formats may need padding; width may be negative now
    if (!left) {
      // if padding with zeros and the first character is a sign then
      // output the sign before padding ("sign" may be ' ')
      if (pad=='0' && (*source=='-' || *source=='+' || *source==' ')) {
        putC(*source); count++;
        width--;    // (may push it negative)
        source++;   // step past sign
        sourlen--;  // and account
        }
      for (; width>sourlen; width--) {putC(pad); count++;}
      }
    // output the real data
    for (; sourlen>0; source++, sourlen--, width--) {putC(*source); count++;}
    // padding might be afterwards
    if (left) for (; width>0; width--) {putC(pad); count++;}
    } // format scan

  return count;
  } // strf

